# If you have come till here, and you want to work on more things like this, reach out:
# research@nimblebox.ai

import os
from time import sleep

import threading

from nbox.jobs import Job

from .. import utils as U
from ..utils import logger
from ..init import nbox_grpc_stub
from ..auth import secret
from ..hyperloop.job_pb2 import Job
from ..hyperloop.dag_pb2 import Node
from ..hyperloop.nbox_ws_pb2 import UpdateRunRequest
from ..messages import rpc, get_current_timestamp

class Tracer:
  def __init__(self, heartbeat_every: int = 60):
    self.heartbeat_every = heartbeat_every
    run_data = secret.get("run") # user should never have "run" on their local
    if run_data == None:
      self.network_tracer = False
      return
    
    init_folder = U.ENVVARS.NBOX_JOB_FOLDER(None)
    if init_folder == None:
      raise RuntimeError("NBOX_JOB_FOLDER not set")
    if not os.path.exists(init_folder):
      raise RuntimeError(f"NBOX_JOB_FOLDER {init_folder} does not exist")

    # get this data from the local secrets file
    self.job_id = run_data["job_id"]
    self.token = run_data["token"]
    with open(U.join(init_folder, "job_proto.msg"), "rb") as f:
      self.job_proto = Job()
      self.job_proto.ParseFromString(f.read())

    self.job_proto.id = self.job_id # because when creating a new job, client does not know the ID
    self.workspace_id = self.job_proto.auth_info.workspace_id
    self.network_tracer = True
    
    logger.debug(f"Username: {self.job_proto.auth_info.username}")
    logger.debug(f"Job Id: {self.job_proto.id}")
    logger.debug(f"Run Id: {self.token}")
    logger.debug(f"Workspace Id: {self.job_proto.auth_info.workspace_id}")
    self.job_proto.status = Job.Status.ACTIVE # automatically first run will

    # start heartbeat in a different thread
    self.thread = threading.Thread(target=self.hearbeat_thread_worker)
    self.thread.start()

  def __call__(self, node: Node, verbose: bool = True):
    if verbose:
      logger.debug(node)
    if not self.network_tracer:
      return
    self.job_proto.dag.flowchart.nodes[node.id].CopyFrom(node) # even if fails we can keep caching this
    updated_at = node.run_status.start if node.run_status.end != None else node.run_status.end
    logger.info(updated_at.ToDatetime())
    rpc(
      nbox_grpc_stub.UpdateRun,
      UpdateRunRequest(token = self.token, job = self.job_proto, updated_at = updated_at),
      f"Could not update job {self.job_proto.id}",
      raise_on_error = False
    )

  def hearbeat_thread_worker(self):
    while True:
      rpc(
        nbox_grpc_stub.UpdateRun,
        UpdateRunRequest(token = self.token, job = self.job_proto, updated_at = get_current_timestamp()),
        "Heartbeat failed",
        raise_on_error = False
      )
      sleep(self.heartbeat_every)

  def stop(self):
    if not self.network_tracer:
      return
    self.thread.join()



# this script is meant to be run to generate latest code for on_ml.py utility functions

import re
import jinja2
import inspect
from textwrap import indent
from typing import Callable, List
from datetime import datetime, timezone

from nbox.framework.model_spec_pb2 import ModelSpec

from nbox import utils as U
from nbox.utils import logger

class MLFrameworkRegistry:
  def __init__(self):
    self.stub_classes = {}
    self.conditional_fns = {}
    self.repr_string = []

    self.packages = {}

  def __repr__(self) -> str:
    return f"< nbox.framework.{self.__class__.__qualname__}\n" + \
      indent("\n".join(self.repr_string), "  ") + \
      "\n>"

  def conditional(
    self,
    framework: str
  ):
    def _wrapper(fn):
      self.conditional_fns[framework] = fn
      return fn
    return _wrapper

  def register_package(
    package_name: str,
    conditional_fn: Callable,
    requirements_txt: List[str],
  ):
    # registers a package and supporting items
    pass

  def register_deserializer(
    package_name: str,
    framework_name: str,
    stub_name: str,
    stub_function_import_string: str,
  ):
    # these are the set of functions that are used to import a model into nbox.Model like
    # nbox.Model.from_xxx(...)

    """
    Ex. Importing from torchscript

    package_name = "torch"
    framework_name = "torchscript"
    stub_name = "TorchScriptLoad"
    stub_function_import_string = "from torch.jit import load"

    nbox.Model.from_torchscript(
      TorchScriptLoad(...), # <-- "load" fn arguments converted to a stub
      NboxOptions(...),     #
    )
    """
    pass

  def register_serializer(
    package_name: str,
    framework_name: str,
    stub_name: str,
    stub_function_import_string: str,
  ) -> ModelSpec:
    # these are the set of functions that are used to export a nbox.Model object to some other format
    # nbox.Model.to_xxx(...)

    """
    Ex: Exporting torch to torchscript

    package_name = "torch"
    framework_name = "torchscript"
    stub_name = "TorchToTorchScript"
    stub_function_import_string = "from torch.jit import trace"

    nbox.Model.to_torchscript(
      TorchToTorchScript(...), # <-- "trace" fn arguments converted to a stub
      NboxOptions(...),        # 
    )
    """
    pass

  def register(
    self,
    stub_name,
    framework: str,
    target: str,
    message_name: str,
    export_fn_import: str,
    target_processor_name: str = None,
    dependencies: List[str] = [],
    ignore_args: List[str] = [],
    verbose = True,
  ):
    def _wrapper(processor_fn: Callable):
      if verbose:
        logger.debug(f"Registering {message_name} and {processor_fn.__name__}")
      
      # import the function if raises import error return None
      export_fn = re.findall(r"import (.*)", export_fn_import)[0]
      if verbose:
        logger.debug(f"export_fn: {export_fn}")
        logger.debug(f"export_fn_import: {export_fn_import}")
        self.repr_string.append(f"{message_name} -> {processor_fn.__name__}")
      try:
        exec(export_fn_import, locals())
      except ImportError:
        logger.error(f"ImportError: {export_fn_import}")
        return None
      export_fn = eval(f"{export_fn}", locals())

      # create the arguments for the template
      out = inspect.getfullargspec(export_fn)
      args = {}
      if out.defaults != None:
        reverse_dict = {}
        for k,v in zip(out.args[::-1], out.defaults[::-1]):
          reverse_dict[k] = v
        back = {k:v for k,v in reversed(reverse_dict.items())}
        for i in range(len(out.args) - len(out.defaults)):
          args[out.args[i]] = None
        args.update(back)
      else:
        for i in range(len(out.args)):
          args[out.args[i]] = None

      arg_strings = []
      final_args = []
      _ignored = {}
      for k,v in args.items():
        if k in ignore_args:
          _ignored[k] = v
          continue
        if re.findall(r"<(.*)>", repr(v)):
          # probably any python object like: <torch.jit.CompilationUnit object at 0x10b682c30>
          _ignored[k] = v
          continue
        if v != None and not isinstance(v, (bool, int, float, str)):
          v = None
        arg_strings.append(f"{k}={v}")
        final_args.append(k)

      template_data = dict(
        dependencies = dependencies,
        framework = framework,
        target = target,
        target_processor_name = target_processor_name,
        
        # message related
        message_name = message_name,
        arg_strings = arg_strings,
        doc = inspect.getdoc(export_fn),
        args = final_args,

        # stub related
        stub_name = stub_name,
        processor_fn = processor_fn.__name__,
      )
      self.stub_classes.setdefault(framework, []).append(template_data)
      return processor_fn
    return _wrapper

  def prepare(self,):
    prepared_frameworks = []
    for frm, template in self.stub_classes.items():
      cond_fn = self.conditional_fns.get(frm, "")
      if cond_fn:
        cond_fn = cond_fn.__name__
        
      data = {
        "messages": [],
        "stubs": [],
        "dependencies": [],
      }
      for template_data in template:
        data["messages"].append({
          "message_name": template_data["message_name"],
          "arg_strings": template_data["arg_strings"],
          "doc": template_data["doc"],
          "args": template_data["args"],
          "dependencies": template_data["dependencies"],
        })
        data["stubs"].append({
          "stub_name": template_data["stub_name"],
          "processor_fn": template_data["processor_fn"],
          "message_name": template_data["message_name"],
          "dependencies": template_data["dependencies"],
          "target": template_data["target"],
          "target_processor_name": template_data["target_processor_name"],
        })
      prepared_frameworks.append((data, frm, cond_fn))
    return prepared_frameworks

# create a register where we can store all the methods
ml_register = MLFrameworkRegistry()

# compiler function
def compile():
  src_file =  U.join(U.folder(U.folder(__file__)), "assets", "ml.jinja")
  trg_file = U.join(U.folder(__file__), "ml.py") # ml.py in this folder
  data = ml_register.prepare()

  # for (data, frm) in data:
  #   print(type(data))
  #   for (data, frm) in data:
  #     print(frm)

  with open(src_file, "r") as src, open(trg_file, "w") as trg:
    trg.write(
      jinja2.Template(src.read()).render(
        timestamp = datetime.now(timezone.utc).isoformat(),
        frameworks = data,
        zip = zip,
    ))



"""
Utility objects and functions.

This has a couple of cool things:

#. ``get_logger``: a master logger for nbox, this can be modified to log through anything
#. ``isthere``: a decorator that checks if a package is installed, if not it raises an error\
  It is more complicated than it needs to be because it is seedling for a way to package\
  functions and code together so that it can be used in a more dynamic way.
#. ``get_files_in_folder``: a function that returns all files in a folder with a certain extension
#. ``fetch``: a function that fetches a url and caches it in ``tempdir`` for faster loading
#. ``get_random_name``: a function that returns a random name, if ``True`` is passed returns\
  an ``uuid4()`` for truly random names :)
#. ``hash_``: a function that returns a hash of any python object, string is accurate, others\
  might be anything, but atleast it returns something.
#. ``folder/join``: to be used in pair, ``join(folder(__file__), "server_temp.jinja")`` means\
  relative path "../other.py". Scattered throughout the codebase, the super generic name will\
  bite me.
#. ``to_pickle/from_pickle``: to be used in pair, ``to_pickle(obj, "path")`` and ``from_pickle("path")``\
  to save and load python objects to disk.
#. ``DBase``: सस्ता-protobuf (cheap protobuf), can be nested and ``get_dict`` will get for all\
  children
#. ``PoolBranch``: Or how to use multiprocessing, but blocked so you don't have to give a shit\
  about it.
"""

############################################################
# This file is d0 meaning that this has no dependencies!
# Do not import anything from rest of nbox here!
############################################################

# this file has bunch of functions that are used everywhere

import os
import io
import sys
import dill
import logging
import hashlib
import requests
import tempfile
import randomname
from uuid import uuid4
from pythonjsonlogger import jsonlogger
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, as_completed

class ENVVARS:
  """
  Single namespace for all environment variables.
  
  #. ``NBOX_LOG_LEVEL``: Logging level for ``nbox``, set ``NBOX_LOG_LEVEL=info|debug|warning"
  #. ``NBOX_JSON_LOG``: Whether to print json-logs, set ``NBOX_JSON_LOG=1``
  #. ``NBOX_JOB_FOLDER``: Folder path for the job, set ``NBOX_JOB_FOLDER=/tmp/nbox/jobs"``
  """
  NBOX_LOG_LEVEL = lambda x: os.getenv("NBOX_LOG_LEVEL", x)
  NBOX_JSON_LOG = lambda x: os.getenv("NBOX_JSON_LOG", x)
  NBOX_JOB_FOLDER = lambda x: os.getenv("NBOX_JOB_FOLDER", x)
  NBOX_NO_AUTH = lambda x: os.getenv("NBOX_NO_AUTH", x)
  NBOX_SSH_NO_HOST_CHECKING = lambda x: os.getenv("NBOX_SSH_NO_HOST_CHECKING", x)


def get_logger():
  logger = logging.getLogger("utils")
  lvl = ENVVARS.NBOX_LOG_LEVEL("info").upper()
  logger.setLevel(getattr(logging, lvl))

  if ENVVARS.NBOX_JSON_LOG(False):
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(jsonlogger.JsonFormatter(
      '%(timestamp)s %(levelname)s %(message)s ',
      timestamp=True
    ))
    logger.addHandler(logHandler)
  else:
    logHandler = logging.StreamHandler()
    logHandler.setFormatter(logging.Formatter(
      '[%(asctime)s] [%(levelname)s] [%(filename)s:%(lineno)d] %(message)s',
      datefmt = "%Y-%m-%dT%H:%M:%S%z"
    ))
    logger.addHandler(logHandler)

  return logger

logger = get_logger() # package wide logger

def log_and_exit(msg, *args, **kwargs):
  # convinience function to avoid tracebacks
  logger.error(msg, *args, **kwargs)
  sys.exit(1)

# lazy_loading/

class Fn:
  # TODO: @yashbonde build this, any arbitrary function
  def __init__(self, fn, requirements = None):
    self.fn = fn
    self.requirements = requirements

  def __repr__(self) -> str:
    return f"<Fn {self.fn.__module__}.{self.fn.__qualname__}>"

  def __call__(self, *args, **kwargs):
    return self.fn(*args, **kwargs)

def isthere(*packages, soft = True):
  """Checks all the packages

  Args:
      soft (bool, optional): If ``False`` raises ``ImportError``. Defaults to True.
  """
  def wrapper(fn):
    _fn_ = Fn(fn, packages)
    def _fn(*args, **kwargs):
      # since we are lazy evaluating this thing, we are checking when the function
      # is actually called. This allows checks not to happen during __init__.
      for package in packages:
        # split the package name to get version number as well, not all will have package
        # version so continue those that have a version
        package = package.split("==")
        if len(package) == 1:
          package_name, package_version = package[0], None
        else:
          package_name, package_version = package
        if package_name in sys.modules:
          # trying to install these using pip will cause issues, so avoid that
          continue

        try:
          module = __import__(package_name)
          if hasattr(module, "__version__") and package_version:
            if module.__version__ != package_version:
              raise ImportError(f"{package_name} version mismatch")
        except ImportError:
          if not soft:
            raise ImportError(f"{package} is not installed, but is required by {fn}")
          # raise a warning, let the modulenotfound exception bubble up
          logger.warning(
            f"{package} is not installed, but is required by {fn.__module__}, some functionality may not work"
          )
      return _fn_(*args, **kwargs)
    return _fn
  return wrapper


# /lazy_loading

# path/

def get_files_in_folder(folder, ext = ["*"]):
  """Get files with ``ext`` in ``folder``"""
  # this method is faster than glob
  import os
  all_paths = []
  _all = "*" in ext # wildcard means everything so speed up

  folder_abs = os.path.abspath(folder)
  for root,_,files in os.walk(folder_abs):
    if _all:
      all_paths.extend([join(root, f) for f in files])
      continue

    for f in files:
      for e in ext:
        if f.endswith(e):
          all_paths.append(os.path.join(root,f))
  return all_paths

def fetch(url, force = False):
  """Fetch and cache a url for faster loading, ``force`` re-downloads"""
  fp = join(tempfile.gettempdir(), hash_(url))
  if os.path.isfile(fp) and os.stat(fp).st_size > 0 and not force:
    with open(fp, "rb") as f:
      dat = f.read()
  else:
    dat = requests.get(url).content
    with open(fp + ".tmp", "wb") as f:
      f.write(dat)
    os.rename(fp + ".tmp", fp)
  return dat

def folder(x):
  """get the folder of this file path"""
  return os.path.split(os.path.abspath(x))[0]

def join(x, *args):
  """convienience function for os.path.join"""
  return os.path.join(x, *args)

def to_pickle(obj, path):
  with open(path, "wb") as f:
    dill.dump(obj, f)

def from_pickle(path):
  with open(path, "rb") as f:
    return dill.load(f)

NBOX_HOME_DIR = os.environ.get("NBOX_HOME_DIR", join(os.path.expanduser("~"), ".nbx"))

# /path

# misc/

def get_random_name(uuid = False):
  """Get a random name, if ``uuid`` is ``True``, return a uuid4"""
  if uuid:
    return str(uuid4())
  return randomname.generate()

def hash_(item, fn="md5"):
  """Hash sting of any item"""
  return getattr(hashlib, fn)(str(item).encode("utf-8")).hexdigest()

# /misc

# datastore/

class DBase:
  def __init__(self, **kwargs):
    for k, v in kwargs.items():
      setattr(self, k, v)

  def get(self, k, v = None):
    return getattr(self, k, v)
  
  def get_dict(self):
    data = {}
    for k in self.__slots__:
      _obj = getattr(self, k, None)
      if _obj == None:
        continue
      if isinstance(_obj, DBase):
        data[k] = _obj.get_dict()
      elif _obj != None and isinstance(_obj, (list, tuple)) and len(_obj) and isinstance(_obj[0], DBase):
        data[k] = [_obj.get_dict() for _obj in _obj]
      else:
        data[k] = _obj
    return data

  def __repr__(self):
    return str(self.get_dict())

# /datastore

# model/

@isthere("PIL", soft = False)
def get_image(file_path_or_url):
  from PIL import Image
  if os.path.exists(file_path_or_url):
    return Image.open(file_path_or_url)
  else:
    return Image.open(io.BytesIO(fetch(file_path_or_url)))

def convert_to_list(x):
  # recursively convert tensors -> list
  import torch
  import numpy as np

  x = x.outputs.detach()

  if isinstance(x, list):
    return x
  if isinstance(x, dict):
    return {k: convert_to_list(v) for k, v in x.items()}
  elif isinstance(x, (torch.Tensor, np.ndarray)):
    x = np.nan_to_num(x, -1.42069)
    return x.tolist()
  else:
    raise Exception("Unknown type: {}".format(type(x)))

# /model


################################################################################
# Parallel
# ========
# There already are many multiprocessing libraries for thread, core, pod, cluster
# but thee classes below are inspired by https://en.wikipedia.org/wiki/Collective_operation
#
# And that is why they are blocking classes, ie. it won't stop till all the tasks
# are completed. For reference please open the link above which has diagrams, there
# nodes can just be threads/cores/... Here is description for each of them:
#
# - Pool: apply same functions on different inputs
# - Branch: apply different functions on different inputs
################################################################################

# pool/

class PoolBranch:
  def __init__(self, mode = "thread", max_workers = 2, _name: str = None):
    """Threading is hard, your brain is not wired to handle parallelism. You are a blocking
    python program. So a blocking function for you. There are some conditions:


    Usage:

    .. code-block:: python

      # define some functions

      def add_zero(x):  return x + 0
      def add_one(x):   return x + 1
      def add_ten(x):   return x + 10
      def add_fifty(x): return x + 50

      all_fn = [add_zero, add_one, add_ten, add_fifty]

      # define some arguments
      args = [(1,), (2,), (3,), (4,),]

      # branching is applying different functions on different inputs
      branch = PoolBranch("thread")
      out = branch(all_fn, *args)

      # pooling is applying same functions on different inputs
      pool = PoolBranch("thread")
      out = pool(add_zero, *args)

    When using ``mode = "process"`` write your code in a function and ensure that the
    function is called from ``__main__ == "__name__"``. From the documentation of ``concurrent.futures``:

      The __main__ module must be importable by worker subprocesses. This means that
      ``ProcessPoolExecutor`` will not work in the interactive interpreter.

    - `StackOverflow <https://stackoverflow.com/questions/27932987/multiprocessing-package-in-interactive-python>`_
    - `Another <https://stackoverflow.com/questions/24466804/multiprocessing-breaks-in-interactive-mode>`_

    .. code-block:: python

      def multiprocess():
        print("MultiProcessing")

        branch = PoolBranch("process")
        out = branch(all_fn, *args)

        pool = PoolBranch("process")
        out = pool(add_zero, *args)

      if __name__ == "__main__":
        multiprocess()

    Args:
      mode (str, optional): There can be multiple pooling strategies across cores, threads,
        k8s, nbx-instances etc.
      max_workers (int, optional): Numbers of workers to use
    """
    self.mode = mode
    self.item_id = -1 # because +1 later
    self.futures = {}
    self._name = _name or get_random_name(True)

    if mode == "thread":
      self.executor = ThreadPoolExecutor(
        max_workers=max_workers,
        thread_name_prefix=self._name
      )
    elif mode == "process":
      self.executor = ProcessPoolExecutor(
        max_workers=max_workers,
      )
    else:
      raise Exception(f"Only 'thread/process' modes are supported")

    logger.debug(f"Starting {mode.upper()}-PoolBranch ({self._name}) with {max_workers} workers")

  def __call__(self, fn, *args):
    """Run any function ``fn`` in parallel, where each argument is a list of arguments to
    pass to ``fn``. Result is returned in the **same order as the input**.

      ..code-block

        if fn is callable:
          thread(fn, a) for a in args -> list of results
        elif fn is list and fn[0] is callable:
          thread(_fn, a) for _fn, a in (fn args) -> list of results
    """
    assert isinstance(args[0], (tuple, list))

    futures = {}
    if isinstance(fn, (list, tuple)) and callable(fn[0]):
      assert len(fn) == len(args), f"Number of functions ({len(fn)}) and arguments ({len(args)}) must be same in branching"
    else:
      assert callable(fn), "fn must be callable in pooling"
      fn = [fn for _ in range(len(args))] # convinience

    self.item_id += len(futures)
    results = {}

    if self.mode in ("thread", "process"):
      for i, (_fn, x) in enumerate(zip(fn, args)):
        futures[self.executor.submit(_fn, *x)] = i # insertion index
      for future in as_completed(futures):
        try:
          result = future.result()
          results[futures[future]] = result # update that index
        except Exception as e:
          logger.error(f"{self.mode} error: {e}")
          raise e

      res = [results[x] for x in range(len(results))]

    return res

# /pool



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Framework(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class KwargsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    PACKAGE_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    METHOD_FIELD_NUMBER: builtins.int
    KWARGS_FIELD_NUMBER: builtins.int
    package: typing.Text
    """name of the package and it's version number"""

    path: typing.Text
    """path to read/write the file"""

    method: typing.Text
    """method used to serialise/deserialise the model"""

    @property
    def kwargs(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """optional parameters for the method, these will be used in case of
        deserialisation but can be provided during encoding as well
        """
        pass
    def __init__(self,
        *,
        package: typing.Text = ...,
        path: typing.Text = ...,
        method: typing.Text = ...,
        kwargs: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["kwargs",b"kwargs","method",b"method","package",b"package","path",b"path"]) -> None: ...
global___Framework = Framework

class Tensor(google.protobuf.message.Message):
    """though the name is a Tensor, it represents any kind of input
    output to the model
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    SHAPE_FIELD_NUMBER: builtins.int
    DIMENSION_NAMES_FIELD_NUMBER: builtins.int
    DTYPE_FIELD_NUMBER: builtins.int
    name: typing.Text
    """name/keyword for this message"""

    @property
    def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """shape: if None assume it is not a tensor"""
        pass
    @property
    def dimension_names(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """what are meanings of shapes like batch, etc."""
        pass
    dtype: typing.Text
    """~numpy datatype of the tensor"""

    def __init__(self,
        *,
        name: typing.Text = ...,
        shape: typing.Optional[typing.Iterable[builtins.int]] = ...,
        dimension_names: typing.Optional[typing.Iterable[typing.Text]] = ...,
        dtype: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["dimension_names",b"dimension_names","dtype",b"dtype","name",b"name","shape",b"shape"]) -> None: ...
global___Tensor = Tensor

class Deployment(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _DeploymentTypes:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _DeploymentTypesEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Deployment._DeploymentTypes.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        NBOX_SERVING: Deployment._DeploymentTypes.ValueType  # 0
        """vanilla nbox serving with fastapi"""

    class DeploymentTypes(_DeploymentTypes, metaclass=_DeploymentTypesEnumTypeWrapper):
        """Deployment type"""
        pass

    NBOX_SERVING: Deployment.DeploymentTypes.ValueType  # 0
    """vanilla nbox serving with fastapi"""


    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    id: typing.Text
    """ID for the deployment"""

    name: typing.Text
    """name of the deployment"""

    workspace_id: typing.Text
    """Workspace ID, if None assume in user workspace"""

    type: global___Deployment.DeploymentTypes.ValueType
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        workspace_id: typing.Text = ...,
        type: global___Deployment.DeploymentTypes.ValueType = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","name",b"name","type",b"type","workspace_id",b"workspace_id"]) -> None: ...
global___Deployment = Deployment

class ModelSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SOURCE_FIELD_NUMBER: builtins.int
    TARGET_FIELD_NUMBER: builtins.int
    FOLDER_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    REQUIREMENTS_FIELD_NUMBER: builtins.int
    EXPORTED_TIME_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    DEPLOY_FIELD_NUMBER: builtins.int
    @property
    def source(self) -> global___Framework:
        """source metadata"""
        pass
    @property
    def target(self) -> global___Framework:
        """target metadata"""
        pass
    folder: typing.Text
    """folder path to be zipped"""

    name: typing.Text
    """name of the model being op-ed"""

    id: typing.Text
    """nbox-id of this model, from this we can connect deploy and jobs"""

    @property
    def requirements(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """leaving a bunch of ids empty for future use for the services
        that we want to provide like train/test/ab/deploy/...

        things to be written in the requirements.txt file so
        something like -f https://... can be added here as well.
        """
        pass
    @property
    def exported_time(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time to creation the export"""
        pass
    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Tensor]:
        """this is the input to the model"""
        pass
    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Tensor]:
        """this is the output of the model"""
        pass
    @property
    def deploy(self) -> global___Deployment:
        """information related to the deploy"""
        pass
    def __init__(self,
        *,
        source: typing.Optional[global___Framework] = ...,
        target: typing.Optional[global___Framework] = ...,
        folder: typing.Text = ...,
        name: typing.Text = ...,
        id: typing.Text = ...,
        requirements: typing.Optional[typing.Iterable[typing.Text]] = ...,
        exported_time: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        inputs: typing.Optional[typing.Iterable[global___Tensor]] = ...,
        outputs: typing.Optional[typing.Iterable[global___Tensor]] = ...,
        deploy: typing.Optional[global___Deployment] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["deploy",b"deploy","exported_time",b"exported_time","source",b"source","target",b"target"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["deploy",b"deploy","exported_time",b"exported_time","folder",b"folder","id",b"id","inputs",b"inputs","name",b"name","outputs",b"outputs","requirements",b"requirements","source",b"source","target",b"target"]) -> None: ...
global___ModelSpec = ModelSpec



"""
This code is used to manage the authentication of the entire ``nbox`` package. For authentication
it will create a ``.nbx`` in the user's home directory (``~/.nbx``, in case of linux) and store
a file called ``secrets.json``. This folder will also contain more information and files that
are used elsewhere as well ex. files generated when takling to any instance.

We have also provided simple built in methods to connect to your cloud provider service as a part
of BYOC (bring your own cloud), they are still work on progress, of you are interested in them,
please raise an issue on Github.

"""
import os
import json
import requests
import webbrowser
from getpass import getpass

from .utils import join, NBOX_HOME_DIR, isthere, logger, ENVVARS

# ------ AWS Auth ------ #

class AWSClient:
  @isthere("boto3", "botocore", soft = False)
  def __init__(self, aws_access_key_id: str, aws_secret_access_key: str, region_name: str):
    """Template for creating your own AWS authentication class.

    EXPERIMENTAL: This is not yet ready for use.

    Args:
        aws_access_key_id (str): AWS access key ID
        aws_secret_access_key (str): AWS secret access key
        region_name (str): AWS region name
    """
    self.aws_access_key_id = aws_access_key_id
    self.aws_secret_access_key = aws_secret_access_key
    self.region_name = region_name

  def get_client(self, service_name: str = "s3", **boto_config_kwargs):
    """Get the client object for the given service

    Args:
        service_name (str): _description_. Defaults to "s3".
    """
    import boto3
    from botocore.client import Config as BotoConfig

    return boto3.client(
      service_name,
      aws_access_key_id=self.aws_access_key_id,
      aws_secret_access_key=self.aws_secret_access_key,
      region_name=self.region_name,
      config = BotoConfig(
        signature_version="s3v4",
        **boto_config_kwargs
      )
    )

# ------ GCP Auth ------ #

class GCPClient:
  @isthere("google-cloud-sdk", "google-cloud-storage", soft = False)
  def __init__(self, project_id: str, credentials_file):
    """Template for creating your own authentication class.

    EXPERIMENTAL: This is not yet ready for use.

    Args:
        project_id (str): GCP project ID
        credentials_file: GCP credentials python object, must support .read() method
    """

    from google.oauth2 import service_account
    
    self.project_id = project_id
    self.credentials_file = credentials_file
    self.creds = service_account.Credentials.from_service_account_file(
      self.credentials_file,
      scopes=["https://www.googleapis.com/auth/cloud-platform"]
    )

  def get_client(self, service_name: str = "storage", **gcp_config_kwargs):
    """Get the client object for the given service

    Args:
        service_name (str): GCP service name
    """
    if service_name == "storage":
      from google.cloud import storage
      return storage.Client(
        project=self.project_id,
        credentials=self.creds,
        **gcp_config_kwargs
      )
    

# ------ Azure Auth ------ #

class AzureClient:
  @isthere("azure-storage-blob", soft = False)
  def __init__(self):
    """
    Microsoft Azure

    EXPERIMENTAL: This is not yet ready for use.
    """
    from azure.storage.blob import BlobServiceClient
    from azure.identity import DefaultAzureCredential

    self.blob_service_client = BlobServiceClient(
      credential=DefaultAzureCredential(),
      endpoint="https://nbox.blob.core.windows.net"
    )

  def get_client(self, service_name = "blob", **azure_config_kwargs):
    if service_name == "blob":
      from azure.storage.blob import BlobClient

      return BlobClient(
        self.blob_service_client,
        **azure_config_kwargs
      )

# ------ OCI Auth ------ #

class OCIClient:
  @isthere("oci", "oci-py", soft = False)
  def __init__(self, config_file):
    """
    Oracle Cloud Infrastructure

    EXPERIMENTAL: This is not yet ready for use.
    """
    from oci.config import from_file
    from oci.signer import Signer

    self.config = from_file(config_file)
    self.signer = Signer(
      tenancy=self.config["tenancy"],
      user=self.config["user"],
      fingerprint=self.config["fingerprint"],
      private_key_file_location=self.config["key_file"]
    )

  def get_client(self, service_name = "object_storage", **oci_config_kwargs):

    if service_name == "object_storage":
      from oci.object_storage.models import CreateBucketDetails
      from oci.object_storage.models import CreateMultipartUploadDetails
      from oci.object_storage.models import Object
      from oci.object_storage.models import UploadPartDetails
      from oci.object_storage.object_storage_client import ObjectStorageClient

      return ObjectStorageClient(
        self.config["user"],
        self.signer,
        **oci_config_kwargs
      )


# ------ Digital Ocean Auth ------ #

class DOClient:
  @isthere("doctl", soft = False)
  def __init__(self, config_file):
    """
    Digital Ocean

    EXPERIMENTAL: This is not yet ready for use.
    """
    from doctl.doctl_client import DictCursor
    from doctl.doctl_client import DoctlClient

    self.doctl_client = DoctlClient(
      config_file=config_file,
      cursor_class=DictCursor
    )
  
  def get_client(self, service_name = "object_storage", **oci_config_kwargs):
    if service_name == "object_storage":
      from doctl.object_storage.object_storage_client import ObjectStorageClient

      return ObjectStorageClient(
        self.doctl_client,
        **oci_config_kwargs
      )

# ------ NBX Auth ------ #

class NBXClient:
  def __init__(self, nbx_url = "https://app.nimblebox.ai"):
    """We try to find the values secrets file in the ``~/.nbx/secrets.json``, if not
    found, we ask the user for the email and direct them to browser.
    """
    os.makedirs(NBOX_HOME_DIR, exist_ok=True)
    fp = join(NBOX_HOME_DIR, "secrets.json")

    # if this is the first time starting this then get things from the nbx-hq
    if not os.path.exists(fp):
      logger.info(f"Ensure that you put the email ID you have signed up with!")
      _secrets_url = f"{nbx_url}/secrets"
      logger.info(f"Opening: {_secrets_url}")
      webbrowser.open(_secrets_url)
      access_token = getpass("Access Token: ")
      
      # Once we have the access token, we can get the secrets
      r = requests.get(f"{nbx_url}/api/v1/user/account_details", headers={"Authorization": f"Bearer {access_token}"})
      try:
        r.raise_for_status()
        username = r.json()["data"]["username"]
        email = r.json()["data"]["email"]
      except:
        username = ""
        email = ""

      # create the objects
      self.secrets = {
        "email": email,
        "access_token": access_token,
        "nbx_url": nbx_url,
        "username": username
      }
      with open(fp, "w") as f:
        f.write(repr(self))
      logger.info("Successfully created secrets!")
    else:
      with open(fp, "r") as f:
        self.secrets = json.load(f)
      logger.debug("Successfully loaded secrets!")

  def __repr__(self):
    return json.dumps(self.secrets, indent=2)

  def get(self, item, default=None):
    return self.secrets.get(item, default)

  def put(self, item, value, persist: bool = False):
    self.secrets[item] = value
    if persist:
      with open(join(NBOX_HOME_DIR, "secrets.json"), "w") as f:
        f.write(repr(self))


def init_secret():
  # add any logic here for creating secrets
  if not ENVVARS.NBOX_NO_AUTH(False):
    return NBXClient()
  return None

secret = init_secret()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import dag_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Job(google.protobuf.message.Message):
    """Job is the definition of a Job ceated by nbox, with a schedule"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _Status:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _StatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Job._Status.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: Job._Status.ValueType  # 0
        NOT_SET: Job._Status.ValueType  # 1
        """when user creates from FE"""

        PREPARING: Job._Status.ValueType  # 2
        """if the run is being prepared"""

        PAUSED: Job._Status.ValueType  # 3
        """if user from FE/nbox said pause"""

        SCHEDULED: Job._Status.ValueType  # 4
        """is a run is scheduled"""

        ACTIVE: Job._Status.ValueType  # 5
        """run is running"""

        ERROR: Job._Status.ValueType  # 6
        """run errored -> paused"""

        COMPLETED: Job._Status.ValueType  # 7
        """last run done and there is nothing scheduled"""

        ARCHIVED: Job._Status.ValueType  # 8
        """when user deletes this job"""

        CANCELLED: Job._Status.ValueType  # 9
        """run status if job is pasued or archive"""

        KILLED: Job._Status.ValueType  # 10
        """run status if job is archived or paused while run is active"""

    class Status(_Status, metaclass=_StatusEnumTypeWrapper):
        pass

    UNKNOWN: Job.Status.ValueType  # 0
    NOT_SET: Job.Status.ValueType  # 1
    """when user creates from FE"""

    PREPARING: Job.Status.ValueType  # 2
    """if the run is being prepared"""

    PAUSED: Job.Status.ValueType  # 3
    """if user from FE/nbox said pause"""

    SCHEDULED: Job.Status.ValueType  # 4
    """is a run is scheduled"""

    ACTIVE: Job.Status.ValueType  # 5
    """run is running"""

    ERROR: Job.Status.ValueType  # 6
    """run errored -> paused"""

    COMPLETED: Job.Status.ValueType  # 7
    """last run done and there is nothing scheduled"""

    ARCHIVED: Job.Status.ValueType  # 8
    """when user deletes this job"""

    CANCELLED: Job.Status.ValueType  # 9
    """run status if job is pasued or archive"""

    KILLED: Job.Status.ValueType  # 10
    """run status if job is archived or paused while run is active"""


    class Code(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        class _Type:
            ValueType = typing.NewType('ValueType', builtins.int)
            V: typing_extensions.TypeAlias = ValueType
        class _TypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Job.Code._Type.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
            NOT_SET: Job.Code._Type.ValueType  # 0
            NBOX: Job.Code._Type.ValueType  # 1
            ZIP: Job.Code._Type.ValueType  # 2
            BINARY: Job.Code._Type.ValueType  # 3
        class Type(_Type, metaclass=_TypeEnumTypeWrapper):
            pass

        NOT_SET: Job.Code.Type.ValueType  # 0
        NBOX: Job.Code.Type.ValueType  # 1
        ZIP: Job.Code.Type.ValueType  # 2
        BINARY: Job.Code.Type.ValueType  # 3

        class S3MetaEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text
            value: typing.Text
            def __init__(self,
                *,
                key: typing.Text = ...,
                value: typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

        S3_URL_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        TYPE_FIELD_NUMBER: builtins.int
        S3_META_FIELD_NUMBER: builtins.int
        s3_url: typing.Text
        size: builtins.int
        type: global___Job.Code.Type.ValueType
        @property
        def s3_meta(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
        def __init__(self,
            *,
            s3_url: typing.Text = ...,
            size: builtins.int = ...,
            type: global___Job.Code.Type.ValueType = ...,
            s3_meta: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["s3_meta",b"s3_meta","s3_url",b"s3_url","size",b"size","type",b"type"]) -> None: ...

    class Schedule(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        START_FIELD_NUMBER: builtins.int
        END_FIELD_NUMBER: builtins.int
        CRON_FIELD_NUMBER: builtins.int
        @property
        def start(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        @property
        def end(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
        cron: typing.Text
        def __init__(self,
            *,
            start: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
            end: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
            cron: typing.Text = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["end",b"end","start",b"start"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["cron",b"cron","end",b"end","start",b"start"]) -> None: ...

    ID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    CODE_FIELD_NUMBER: builtins.int
    RESOURCE_FIELD_NUMBER: builtins.int
    AUTH_INFO_FIELD_NUMBER: builtins.int
    SCHEDULE_FIELD_NUMBER: builtins.int
    DAG_FIELD_NUMBER: builtins.int
    PAUSED_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    K8SINFO_FIELD_NUMBER: builtins.int
    id: typing.Text
    """unique id of the job"""

    name: typing.Text
    """user defined name of the job
    // user provided description of the job
    string description = 3;
    """

    @property
    def created_at(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """time created"""
        pass
    @property
    def code(self) -> global___Job.Code: ...
    @property
    def resource(self) -> global___Resource: ...
    @property
    def auth_info(self) -> global___NBXAuthInfo:
        """some blank space for future fields

        """
        pass
    @property
    def schedule(self) -> global___Job.Schedule:
        """schedule of the job"""
        pass
    @property
    def dag(self) -> dag_pb2.DAG:
        """nbox created DAG information
        get this from dag.proto
        """
        pass
    paused: builtins.bool
    """if the job is/has to be paused"""

    status: global___Job.Status.ValueType
    """Job Status, as used within WS"""

    @property
    def k8sInfo(self) -> global___K8sInfo: ...
    def __init__(self,
        *,
        id: typing.Text = ...,
        name: typing.Text = ...,
        created_at: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        code: typing.Optional[global___Job.Code] = ...,
        resource: typing.Optional[global___Resource] = ...,
        auth_info: typing.Optional[global___NBXAuthInfo] = ...,
        schedule: typing.Optional[global___Job.Schedule] = ...,
        dag: typing.Optional[dag_pb2.DAG] = ...,
        paused: builtins.bool = ...,
        status: global___Job.Status.ValueType = ...,
        k8sInfo: typing.Optional[global___K8sInfo] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","code",b"code","created_at",b"created_at","dag",b"dag","k8sInfo",b"k8sInfo","resource",b"resource","schedule",b"schedule"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["auth_info",b"auth_info","code",b"code","created_at",b"created_at","dag",b"dag","id",b"id","k8sInfo",b"k8sInfo","name",b"name","paused",b"paused","resource",b"resource","schedule",b"schedule","status",b"status"]) -> None: ...
global___Job = Job

class NBXAuthInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    USERNAME_FIELD_NUMBER: builtins.int
    WORKSPACE_ID_FIELD_NUMBER: builtins.int
    ACCESS_TOKEN_FIELD_NUMBER: builtins.int
    REFRESH_TOKEN_FIELD_NUMBER: builtins.int
    ACCESS_KEY_FIELD_NUMBER: builtins.int
    username: typing.Text
    workspace_id: typing.Text
    """Workspace ID"""

    access_token: typing.Text
    refresh_token: typing.Text
    access_key: typing.Text
    """NBX Access Key for use with Jobs"""

    def __init__(self,
        *,
        username: typing.Text = ...,
        workspace_id: typing.Text = ...,
        access_token: typing.Text = ...,
        refresh_token: typing.Text = ...,
        access_key: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["access_key",b"access_key","access_token",b"access_token","refresh_token",b"refresh_token","username",b"username","workspace_id",b"workspace_id"]) -> None: ...
global___NBXAuthInfo = NBXAuthInfo

class K8sInfo(google.protobuf.message.Message):
    """Some k8s cluster resource info"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CRON_JOB_FIELD_NUMBER: builtins.int
    PODS_FIELD_NUMBER: builtins.int
    cron_job: typing.Text
    """Name of CronJob obj in k8s"""

    @property
    def pods(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
        """Names of pods in k8s"""
        pass
    def __init__(self,
        *,
        cron_job: typing.Text = ...,
        pods: typing.Optional[typing.Iterable[typing.Text]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["cron_job",b"cron_job","pods",b"pods"]) -> None: ...
global___K8sInfo = K8sInfo

class Resource(google.protobuf.message.Message):
    """Resources required by Job"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CPU_FIELD_NUMBER: builtins.int
    MEMORY_FIELD_NUMBER: builtins.int
    GPU_FIELD_NUMBER: builtins.int
    GPU_COUNT_FIELD_NUMBER: builtins.int
    DISK_SIZE_FIELD_NUMBER: builtins.int
    TIMEOUT_FIELD_NUMBER: builtins.int
    MAX_RETRIES_FIELD_NUMBER: builtins.int
    cpu: typing.Text
    """CPU in milli-core"""

    memory: typing.Text
    """Memory in MiB"""

    gpu: typing.Text
    """GPU Type"""

    gpu_count: typing.Text
    """Number of GPUs as a fraction"""

    disk_size: typing.Text
    """in GiBs"""

    timeout: builtins.int
    """in seconds. Note this is inclusive of jobs and init containers, enforced by k8s, regardless of `max_retries` set"""

    max_retries: builtins.int
    """max number of retries for a job"""

    def __init__(self,
        *,
        cpu: typing.Text = ...,
        memory: typing.Text = ...,
        gpu: typing.Text = ...,
        gpu_count: typing.Text = ...,
        disk_size: typing.Text = ...,
        timeout: builtins.int = ...,
        max_retries: builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["cpu",b"cpu","disk_size",b"disk_size","gpu",b"gpu","gpu_count",b"gpu_count","max_retries",b"max_retries","memory",b"memory","timeout",b"timeout"]) -> None: ...
global___Resource = Resource



"""
This file loads first and is responsible for setting up all the global networking items.

..

  Build systems are an extremely important detail of any software project. When they work,
  they can feel like magic: you execute a command, and after a series of potentially
  complicated steps, a working binary (or other artifact) is produced! When they don't work,
  they can feel like confusing, non-transparent roadblocks that you wish didn't exist. This
  is typical for any powerful tool: magic or a headache depending on the day and the task.
"""

import grpc
import requests

from .auth import secret
from .utils import logger
from .subway import Sub30
from .hyperloop.nbox_ws_pb2_grpc import WSJobServiceStub
from . import version as V

def get_stub() -> WSJobServiceStub:
  """Create a gRPC stub with the NBX Webserver, this will initialise ``nbox_grpc_stub``
  object which is globally accesible as ``nbox.nbox_grpc_stub``. If you find yourself
  using this function, you might want to reconsider your design."""
  token_cred = grpc.access_token_call_credentials(secret.get("access_token"))
  ssl_creds = grpc.ssl_channel_credentials()
  creds = grpc.composite_channel_credentials(ssl_creds, token_cred)
  channel = grpc.secure_channel(secret.get("nbx_url").replace("https://", "dns:/")+":443", creds)
  stub = WSJobServiceStub(channel)
  future = grpc.channel_ready_future(channel)
  future.add_done_callback(lambda _: logger.info(f"NBX gRPC server is ready"))
  logger.info(f"Connected using stub: {stub.__class__.__name__}")
  return stub

def create_webserver_subway(version: str = "v1", session: requests.Session = None) -> Sub30:
  """Create a Subway object for the NBX Webserver. This is a wrapper around the
  OpenAPI spec plublished by NBX Webserver. It loads the JSON object in ``Sub30``
  which allows accesing REST APIs with python "." (dot) notation. If you find yourself
  using this function, you might want to reconsider your design.

  Returns:
    Sub30: A Subway object for the NBX Webserver.
  """
  _version_specific_url = secret.get("nbx_url") + f"/api/{version}"
  session = session if session != None else nbox_session # select correct session
  r = session.get(_version_specific_url + "/openapi.json")
  try:
    r.raise_for_status()
  except Exception as e:
    logger.error(f"Could not connect to webserver at {secret.get('nbx_url')}")
    logger.error(e)
    return None
  out = Sub30(_version_specific_url, r.json(), session)
  logger.info(f"Connected to webserver at {out}")
  return out


# common networking items that will be used everywhere
nbox_session = requests.Session()
nbox_session.headers.update({"Authorization": f"Bearer {secret.get('access_token')}"})
nbox_grpc_stub: WSJobServiceStub  = get_stub()
nbox_ws_v1: Sub30 = create_webserver_subway(version = "v1", session = nbox_session)

# TODO: @yashbonde: raise deprecation warning for version
# raise_old_version_warning(V._major, V._minor, V._patch)



"""
This is CLI for ``nbox``, it is meant to be as simple to use as possible.

..code-block::

  nbx
  ├── tunnel
  ├── home
  ├── compile
  ├── get
  ├── jobs
  │   [staticmethods]
  │   ├── new
  │   ├── status
  │   [initialisation]
  │       ├── id
  │       └── workspace_id
  │   ├── change_schedule
  │   ├──logs
  │   ├──delete
  │   ├──refresh
  │   ├──trigger
  │   ├──pause
  │   └──resume
  └── build (Instance)
      [staticmethods]
      ├── new
      ├── status
      [initialisation]
          ├── i
          ├── workspace_id
          └── cs_endpoint
      ├── refresh
      ├── start
      ├── stop
      ├── delete
      ├── run_py
      └── __call__

"""

import os
import sys
import fire
from json import dumps

from nbox.auth import init_secret
from .jobs import Job, new_model
from .instance import Instance
from .sub_utils import ssh
from .framework.autogen import compile
from .init import nbox_ws_v1
from nbox import utils as U

def open_home():
  """Open current NBX platform"""
  from .auth import secret
  import webbrowser
  webbrowser.open(secret.get("nbx_url"))


def get(api_end: str, **kwargs):
  """Get any command 

  Args:
    api_end (str): something like '/v1/api'
  """
  api_end = api_end.strip("/")
  if nbox_ws_v1 == None:
    raise RuntimeError("Not connected to NimbleBox.ai webserver")
  out = nbox_ws_v1
  for k in api_end.split("/"):
    out = getattr(out, k)
  res = out(_method = "get", **kwargs)
  sys.stdout.write(dumps(res))
  sys.stdout.flush()

def login():
  init_secret()


class ServeCLINamespace(object):
  new = staticmethod(new_model)

  def __call__(
    self,
    init_folder,
    deployment_id_or_name: str,
    workspace_id: str = None,
    wait_for_deployment: bool = True,
  ):
    from nbox import logger

    folder = os.path.abspath(init_folder)
    sys.path.append(folder)
    logger.info(f"Compiling serving: {folder}")

    # get the items from users code
    from nbx_user import get_op
    try:
      from nbx_user import get_resource
      resource = get_resource()
    except ImportError:
      # old version problems
      resource = None

    from nbox import Operator

    op: Operator = get_op()
    return op.serve(
      init_folder=init_folder,
      deployment_id_or_name = deployment_id_or_name,
      workspace_id = workspace_id,
      resource = resource,
      wait_for_deployment = wait_for_deployment,
    )

def get_dict():
  from .operator import Operator
  NBX = dict(
    login = login                  , # nbox login
    tunnel = ssh.tunnel            , # nbox tunnel
    home = open_home               , # nbox home
    compile = compile              , # nbox compile: internal for autogen code
    get = get                      , # nbox get "/workspace/88fn83/projects"
    #
    build = Instance               , # nbox build
    jobs = Job                     , # nbox jobs
    serve = ServeCLINamespace      , # nbox serve
    unzip = Operator.unzip         , # nbox unzip (convinience)
  )
  return NBX

def main():
  data = get_dict()
  fire.Fire(data)

if __name__ == "__main__":
  main()



"""
This file has the code to run ``nbox.Model`` which is top of the stack class. It can load
any arbitrary model and run it, deploy it (if possible to export), train it (WIP). It has gone
through it's fare share of refactoring and is now much simpler, in terms of functions while
providing very abstract value.

tl;dr
-----

Underneath this system uses a simple and effective strategy. All the supported methods for
all the frameworks (tensoflow, pytorch, etc) are in a registery, whenever the user sends
any model, it is matched and all functions for that framework (eg. ``sklearn_to_pickle``)
are loaded. This then provides first class support eg. ``delay_forecaster.torch_to_onnx()``,
combined with the protobuf style message stub system this provides all configurations to
user and leaves us incharge of simply executing it.

Story Mode
----------

Let's just say your organisation is working on project for predicting "User Fraud Prevention",
and you already have a working baseline in scikit learn. You have already trained the model
and now want to deploy it to production.

.. code-block:: python

  # get your baseline model
  sklearn_model: 'SklearnModel' = get_skl_model(...)
  fraud_scoring: Model = Model(model = sklearn_model, method = "predict_proba")

  # sklearn models consume numpy tesors so we send them those
  full_matrix: np.ndarray = get_test_data()
  y_pred: np.ndarray = stock_prediction(x = full_matrix[:, :-1])

Before deploying you can write simple functions that can be used to preprocess and postprocess
the incoming data. Your function consumes``UserProto`` or a ``Dict`` that can be parsed to
proto. So you create a new model like this:

.. code-block:: python

  def _pre(user: Union[UserProto, Dict]) -> Dict:
    if isinstance(str):
      user = dict_to_message(user, UserProto) # deployment works on REST so load dict to proto

    # maybe you want to update your DB
    sql.execute(f'INSERT INTO users (id, name) VALUES ({user.id}, {user.name})')

    ... # your business logic here

    return {"X": user.feature_vector.numpy(),} # NOTE: returns dict

  fraud_scoring = Model(
    model = sklearn_model,
    method = "predict_proba",
    pre = _pre
  )

  # You can now pass both proto and dict to the model
  user: UserProto = get_user_from_db()
  prediction: Dict = fraud_scoring(user)

If you want to add a post processing (eg. here I am just converting the prediction to a list):

.. code-block:: python

  def _post(pred: Dict) -> Dict:
    # maybe you want to update your DB
    sql.execute(f'UPDATE users SET fraud_score = {pred["fraud_score"]} WHERE id = {pred["user_id"]}')

    ... # your business logic here

    return {k:v.tolist() for k,v in pred} # NOTE: returns json serializable dict

  fraud_scoring = Model(
    model = sklearn_model,
    method = "predict_proba",
    pre = _pre,
    post = _post
  )

When you want to deploy the model, you will need to export it to some format. You don't have to
search through documentation to find out supported models, they are already in the object as
``extra_fns`` attribute (``fraud_scoring.extra_fns``). You chose to go the reliable way of pickling
and unpickling.

.. code-block:: python

  from nbox.framework import SklearnToPickle, NboxOptions, ModelSpec

  # get a protobuf dataclass for the spec
  model_spec: ModelSpec = fraud_scoring.sklearn_to_pickle(
    SklearnToPickle( ... ), # fill values based on your preferance
    NboxOptions( ... ) # fill values based on your preferance
  )

You have the ``ModelSpec`` object and you can now deploy it. All the information is packaged
in a nice serving that will be put on K8s.

.. code-block:: python

  url, key = fraud_scoring.deploy(
    model_spec = model_spec,
    ... # other deployment args
  )

Here's documentation to the code:
"""

import os
import tarfile
from typing import Any, Dict, Union
from types import SimpleNamespace


from . import utils as U
from .utils import logger
from .init import nbox_ws_v1
from .network import deploy_serving
from .instance import Instance
from .subway import NboxModelSubway
from .framework import get_model_functions
from .framework import ModelSpec, Deployment, NboxOptions
from .messages import message_to_json, dict_to_message

# model/

class Model:
  def __init__(
    self,
    model: Any,
    method: str = None,
    pre: callable = None,
    post: callable = None,
    model_spec: ModelSpec = None,
    verbose: bool = False
  ):
    """Top of the stack Model class.

    Args:
      model (Any): First input to the model
      method (str, optional): If there is a specific function to call for the model.
        if ``None`` then ``__call__`` will be made else ``model.method`` will be called.
        Defaults to None.
      pre (callable): preprocessing function to be applied to the ``input_object``
      post (callable): postprocessing function to be applied to the model prediction
      model_spec (ModelSpec, optional): ModelSpec object to be used for the model.
      verbose (bool, optional): If true provides detailed prints. Defaults to False.
    """

    if model == NboxModelSubway:
      self.model = NboxModelSubway(model)
      method = None
    else:
      self.model = model
    self.method = method
    self.forward_fn = self.model if method == None else getattr(self.model, method)
    self.pre = pre if pre != None else lambda x: x
    self.post = post if post != None else lambda x: x
    self.model_spec = model_spec
    self.verbose = verbose

    self.extra_fns: Dict = get_model_functions(self.model)
    for (fn_name, fn_meta) in self.extra_fns.items():
      fn, _ = fn_meta
      setattr(self, fn_name, fn)
      logger.debug(f"Adding {fn_name}")

  ################################################################################
  # Utility functions
  ################################################################################

  def __repr__(self):
    return f"<nbox.Model: {self.model} >"

  def __dir__(self):
    # be careful: https://docs.python.org/3/library/functions.html#dir
    return [
      # core functions
      "__init__",
      "__repr__",
      "__call__",
      "deserialise",
      "deploy",
      "train_on_instance",

      # attributes
      "model",
      "method",
      "forward_fn",
      "pre",
      "post",
      "model_spec",
      "verbose",
      "extra_fns",

      # framework related functions
      *tuple(self.extra_fns.keys())
    ]

  ################################################################################
  # Functions here are the services that NBX provides
  ################################################################################

  def __call__(self, input_object) -> Any:
    r"""Call is the most important UI/UX. The ``input_object`` can be anything from
    a tensor, an image file, filepath as string, string and is processed by ``pre`` function.

    The entire purpose of this package is to make ML chill.

    Args:
      input_object (Any): input to be processed
    """
    pre_out = self.pre(input_object) # pre processing output
    model_out = self.forward_fn(**pre_out) # model prediction
    post_out = self.post(model_out) # post processing output
    return post_out

  def export_model(self) -> ModelSpec:
    raise NotImplementedError("Define your export function here")

  @classmethod
  def deserialise(cls, model_spec: Union[ModelSpec, Dict], folder) -> 'Model':
    """Load ``ModelSpec`` and ``folder`` with the files in it and return a ``Model`` object.
    
    Args:
      model_spec (Union[ModelSpec, Dict]): ModelSpec object or dictionary of the model_spec
      folder (str): folder where the model files are stored
    """
    if isinstance(model_spec, dict):
      _model_spec = ModelSpec()
      model_spec = dict_to_message(model_spec, _model_spec)
    logger.info(f"{model_spec}")
    init_data = U.from_pickle(U.join(folder, "model.extras.pkl"))

    # now need to load the model from the serialised object
    _class, _method = model_spec.target.method.split(".")
    exec(f"from .framework.ml import {_class}")
    _module = eval(f"{_class}")
    loader, options_cls = _module._METHODS.get(_method)
    m0 = loader(
      user_options = options_cls(),
      nbox_options = NboxOptions(model_name = model_spec.name, folder = folder, create_folder = False)
    )
    return cls(m0, method = init_data["method"], pre = init_data["pre"], post = init_data["post"], model_spec = model_spec)

  def deploy(
    self,
    model_spec: ModelSpec,
    deployment_id_or_name: str = None,
    workspace_id: str = None,
    wait_for_deployment=True,
    *,
    _unittest = False
  ):
    """Serve your model on NBX-Deploy `read more <https://nimbleboxai.github.io/nbox/nbox.model.html>`_

    Args:
      model_spec (nbox.framework.ModelSpec): ModelSpec object
      deployment_id_or_name (str, optional): Deployment information through ID or name, if not
        provided will create a new deployment group with the given name
      workspace_id (str, optional): Workspace ID to deploy the model to. If not provided
        will use the personal workspace.
      wait_for_deployment (bool, optional): Block thread till deployment to be ready.
    """
    if workspace_id == None:
      stub_all_depl = nbox_ws_v1.user.deployments
    else:
      stub_all_depl = nbox_ws_v1.workspace.u(workspace_id).deployments
    logger.debug(f"deployments stub: {stub_all_depl}")

    _deploy_proto = Deployment(
      workspace_id = workspace_id,
      type = Deployment.DeploymentTypes.NBOX_SERVING # ignored for now
    )

    deployments = list(filter(
      lambda x: x["deployment_id"] == deployment_id_or_name or x["deployment_name"] == deployment_id_or_name,
      stub_all_depl()["data"]
    ))
    if len(deployments) == 0:
      logger.warning(f"No deployment found with id '{deployment_id_or_name}', creating one with same name")
      _deploy_proto.name = deployment_id_or_name
    elif len(deployments) > 1:
      raise ValueError(f"Multiple deployments found for '{deployment_id_or_name}', try passing ID")
    else:
      data = deployments[0]
      _deploy_proto.id = data["deployment_id"]
      _deploy_proto.name = data["deployment_name"]

    # update model spec with deployment related information
    model_spec.deploy.CopyFrom(_deploy_proto)

    # pack everything nicely
    folder = model_spec.folder
    extras = U.join(folder, f"model.extras.pkl")
    logger.info(f"Writing model.extras: {extras}")
    U.to_pickle({"pre": self.pre, "post": self.post, "method": self.method}, extras)

    req = U.join(folder, "requirements.txt")
    with open(req, "w") as f:
      logger.info(f"Writing the requirements file: {req}")
      f.write("\n".join(model_spec.requirements))

    meta_path = U.join(folder, f"nbox_config.json")
    with open(meta_path, "w") as f:
      logger.info(f"Writing nbox.meta: {meta_path}")
      f.write(message_to_json(model_spec))

    nbx_path = U.join(folder, f"{folder}.nbox")
    all_files = U.get_files_in_folder(folder)
    with tarfile.open(nbx_path, "w|gz") as tar:
      logger.info(f"Writing: {nbx_path}")
      for path in all_files:
        tar.add(path, arcname = os.path.basename(path))
        logger.debug(f"Removed {path}")

    if _unittest:
      # returns the minimum information needed to deserialise the model
      return model_spec, folder

    # OCD baby!
    return deploy_serving(
      export_model_path=nbx_path,
      stub_all_depl=stub_all_depl,
      model_spec=model_spec,
      wait_for_deployment=wait_for_deployment,
    )

  @staticmethod
  def train_on_instance(
    instance: Instance,
    serialised_fn: callable,
    train_fn: callable,
    other_args: tuple = (), # any other arguments to be passed to the train_fn
    target_folder: str = "/", # anything after /project folder
    shutdown_once_done: bool = False,
    *,
    _unit_test: bool = False,
  ):
    """Train this model on an NBX-Build Instance. Though this function is generic enough to execute
    any arbitrary code, this is built primarily for internal use.

    EXPERIMENTAL: FEATURES MIGHT BREAK

    Args:
      instance (Instance): Instance to train the model on
      serialised_fn (callable): path to the serialised tar file
      train_fn (callable): pure function that trains the model
      other_args (Any, optional): any other arguments to be passed to the train_fn
      target_folder (str, optional): folder on the ``instance`` to run this program in,
        will run in folder ``/project/{target_folder}/``
      shutdown_once_done (bool, optional): if true, shutdown the instance once training is done.
    """

    assert instance.status == "RUNNING", f"Instance {instance.id} is not running"
    all_files, nbx_path = serialised_fn(_do_tar = False, _unit_test = _unit_test)

    train_fn_path = U.join(U.folder(nbx_path), "train_fn.dill")
    logger.debug(f"Train function saved at {train_fn_path}")
    U.to_pickle(SimpleNamespace(train_fn = train_fn, args = other_args), train_fn_path)
    all_files.append(train_fn_path)

    logger.debug(f"Creating nbox zip: {nbx_path}")
    with tarfile.open(nbx_path, "w|gz") as tar:
      for path in all_files:
        tar.add(path, arcname = os.path.basename(path))
        # os.remove(path)
        logger.debug(f"Removed {path}")

    run_folder = f"/project/{target_folder}/"

    instance.mv(nbx_path, run_folder)
    instance("cd {}; python3 -m nbox.train_fn".format(run_folder))
    instance.mv(
      U.join(U.folder(__file__), "assets", "train_fn.jinja"),
      U.join(run_folder, "run.py")
    )

    pid = instance(U.join(run_folder, "run.py"))
    instance.stream_logs(pid)

    if shutdown_once_done:
      instance.stop()

  @staticmethod
  def train_on_job():
    """Train this model on NBX-Jobs. This same experience can be given by creating a new job
    that then can then be populated how we currently create a new job using NBX-Jobs CLI
    ```
    nbx jobs new --help
    ```
    So should we really add this, is the question!
    
    EXPERIMENTAL: FEATURES MIGHT BREAK
    """
    pass


def magic_model():
  from .framework.on_ml import _get_torch_model, _get_sklearn_model, _get_default_forward
  for fn in [_get_torch_model, _get_sklearn_model, _get_default_forward]:
    try:
      model_kwargs, sample_input = fn()
    except ImportError as e:
      pass

  return Model(**model_kwargs), sample_input



from nbox import Operator
from nbox.instance import Instance
from nbox.utils import PoolBranch

class NboxInstanceStartOperator(Operator):
  def __init__(self, instances):
    """Starts multiple instances on nbox in a blocking fashion"""
    super().__init__()
    if not isinstance(instances, list):
      instances = [instances]
    assert instances[0].__class__ == Instance, "instances must be of type nbox.Instance"
    self.instances = instances
    self.pool = PoolBranch("thread", len(instances), _name = "instance_starter")

  def forward(self):
    self.pool(
      lambda instance: instance.start(cpu_only = True),
      self.instances
    )

class NboxModelDeployOperator(Operator):
  def __init__(self, model_name, model_path, model_weights, model_labels):
    """Simple Operator that wraps the deploy function of ``Model``"""
    super().__init__()
    self.model_name = model_name
    self.model_path = model_path
    self.model_weights = model_weights
    self.model_labels = model_labels

  def forward(self, name):
    from nbox.model import Model

    return Model(
      self.model_name,
      self.model_path,
      self.model_weights,
      self.model_labels,
    ).deploy(name)

class NboxWaitTillJIDComplete(Operator):
  def __init__(self, instance, jid):
    """Blocks threads while a certain PID is complete on Instance"""
    super().__init__()
    self.instance = instance
    self.jid = jid

  def forward(self, poll_interval = 5):
    status = self.instance(self.jid)
    if status == "done":
      return None
    elif status == "error-done":
      raise Exception("Job {} failed".format(self.jid))
    elif status == "running":
      from time import sleep
      while status == "running":
        sleep(poll_interval)
        status = self.instance(self.jid)
      if status == "done":
        return None
      elif status == "error-done":
        raise Exception("Job {} failed".format(self.jid))

class NboxInstanceMv(Operator):
  def __init__(self, i: str, workspace_id: str) -> None:
    super().__init__()
    self.build = Instance(i = i, workspace_id = workspace_id)
    self.start_build = NboxInstanceStartOperator(self.build)
  
  def forward(self, src: str, dst: str, force: bool = False) -> None:
    if not self.build.is_running():
      self.start_build()
    resp = self.build.mv(src, dst, force = force)
    return resp



# In case of nbox which handles all kinds of weird paths, initialisation is important.
# We are defining init.py that starts the loading sequence

from .init import nbox_grpc_stub, nbox_session, nbox_ws_v1
from .utils import logger
from .model import Model
from .load import load, PRETRAINED_MODELS
from .jobs import Job
from .instance import Instance
from .auth import AWSClient, GCPClient, OCIClient, DOClient, AzureClient
from .operator import Operator
from .subway import Sub30
from .version import __version__



"""
ML Autogen
==========

This is an autogenerated file DO NOT TOUCH, it's sole intention is to provide IDEs
a way to index functions and thereby supporting the users with autocompletion.

Generated by: autogen.py
Complements: on_ml.py

Time of generation (UTC): 2022-03-21T07:12:52.019635
"""
# -----------------------------------------------------------------------------------

# The classes here are designed like gRPC stubs, templates that can be indexed by the
# IDE and intuitively used by the users by just filling the variable in it.

from typing import Union, Any
from google.protobuf.timestamp_pb2 import Timestamp

from .model_spec_pb2 import ModelSpec, Framework
from .on_ml import NboxOptions

# --------- torch --------- #

class TorchToTorchscript:
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def __init__(
    self,
    func=None,
    example_inputs=None,
    optimize=None,
    check_trace=True,
    check_inputs=None,
    check_tolerance=1e-05,
    strict=True,
    _force_outplace=False,
    _module_class=None,
    **kwargs
  ):
    '''AUTOGENERATED by ``nbox``, some arguments might be missing, USE ``kwargs`` CAREFULLY!

    Trace a function and return an executable  or :class:`ScriptFunction`
    that will be optimized using just-in-time compilation. Tracing is ideal for
    code that operates only on ``Tensor``\s and lists, dictionaries, and
    tuples of ``Tensor``\s.

    Using `torch.jit.trace` and `torch.jit.trace_module`, you can turn an
    existing module or Python function into a TorchScript
    :class:`ScriptFunction` or :class:`ScriptModule`. You must provide example
    inputs, and we run the function, recording the operations performed on all
    the tensors.

    * The resulting recording of a standalone function produces `ScriptFunction`.
    * The resulting recording of `nn.Module.forward` or `nn.Module` produces
      `ScriptModule`.

    This module also contains any parameters that the original
    module had as well.

    Warning:
        Tracing only correctly records functions and modules which are not data
        dependent (e.g., do not have conditionals on data in tensors) and do not have
        any untracked external dependencies (e.g., perform input/output or
        access global variables). Tracing only records operations done when the given
        function is run on the given tensors. Therefore, the returned
        `ScriptModule` will always run the same traced graph on any input. This
        has some important implications when your module is expected to run
        different sets of operations, depending on the input and/or the module
        state. For example,

        * Tracing will not record any control-flow like if-statements or loops.
          When this control-flow is constant across your module, this is fine
          and it often inlines the control-flow decisions. But sometimes the
          control-flow is actually part of the model itself. For instance, a
          recurrent network is a loop over the (possibly dynamic) length of an
          input sequence.
        * In the returned :class:`ScriptModule`, operations that have different
          behaviors in ``training`` and ``eval`` modes will always behave as if
          it is in the mode it was in during tracing, no matter which mode the
          `ScriptModule` is in.

        In cases like these, tracing would not be appropriate and
        :func:`scripting <torch.jit.script>` is a better choice. If you trace
        such models, you may silently get incorrect results on subsequent
        invocations of the model. The tracer will try to emit warnings when
        doing something that may cause an incorrect trace to be produced.

    Args:
        func (callable or torch.nn.Module):  A Python function or `torch.nn.Module`
            that will be run with `example_inputs`. `func` arguments and return
            values  must be tensors or (possibly nested) tuples that contain
            tensors. When a module is passed `torch.jit.trace`, only the
            ``forward`` method is run and traced (see :func:`torch.jit.trace
            <torch.jit.trace_module>` for details).
        example_inputs (tuple or torch.Tensor):  A tuple of example inputs that
            will be passed to the function while tracing. The resulting trace
            can be run with inputs of different types and shapes assuming the
            traced operations support those types and shapes. `example_inputs`
            may also be a single Tensor in which case it is automatically
            wrapped in a tuple.

    Keyword arguments:
        check_trace (``bool``, optional): Check if the same inputs run through
            traced code produce the same outputs. Default: ``True``. You might want
            to disable this if, for example, your network contains non-
            deterministic ops or if you are sure that the network is correct despite
            a checker failure.

        check_inputs (list of tuples, optional): A list of tuples of input
            arguments that should be used to check the trace against what is
            expected. Each tuple is equivalent to a set of input arguments that
            would be specified in ``example_inputs``. For best results, pass in
            a set of checking inputs representative of the space of shapes and
            types of inputs you expect the network to see.  If not specified,
            the original ``example_inputs`` are used for checking
        check_tolerance (float, optional): Floating-point comparison tolerance
            to use in the checker procedure.  This can be used to relax the
            checker strictness in the event that results diverge numerically
            for a known reason, such as operator fusion.
        strict (``bool``, optional): run the tracer in a strict mode or not
            (default: ``True``). Only turn this off when you want the tracer to
            record your mutable container types (currently ``list``/``dict``)
            and you are sure that the container you are using in your
            problem is a ``constant`` structure and does not get used as
            control flow (if, for) conditions.

    Returns:
        If `func` is `nn.Module` or ``forward`` of `nn.Module`, `trace` returns
        a :class:`ScriptModule` object with a single ``forward`` method
        containing the traced code.  The returned `ScriptModule` will
        have the same set of sub-modules and parameters as the original
        ``nn.Module``.  If ``func`` is a standalone function, ``trace``
        returns `ScriptFunction`.

    Example (tracing a function):

    .. testcode::

        import torch

        def foo(x, y):
            return 2 * x + y

        # Run `foo` with the provided inputs and record the tensor operations
        traced_foo = torch.jit.trace(foo, (torch.rand(3), torch.rand(3)))

        # `traced_foo` can now be run with the TorchScript interpreter or saved
        # and loaded in a Python-free environment

    Example (tracing an existing module)::

        import torch
        import torch.nn as nn

        class Net(nn.Module):
            def __init__(self):
                super(Net, self).__init__()
                self.conv = nn.Conv2d(1, 1, 3)

            def forward(self, x):
                return self.conv(x)

        n = Net()
        example_weight = torch.rand(1, 1, 3, 3)
        example_forward_input = torch.rand(1, 1, 3, 3)

        # Trace a specific method and construct `ScriptModule` with
        # a single `forward` method
        module = torch.jit.trace(n.forward, example_forward_input)

        # Trace a module (implicitly traces `forward`) and construct a
        # `ScriptModule` with a single `forward` method
        module = torch.jit.trace(n, example_forward_input)
    '''
    self.func=func
    self.example_inputs=example_inputs
    self.optimize=optimize
    self.check_trace=check_trace
    self.check_inputs=check_inputs
    self.check_tolerance=check_tolerance
    self.strict=strict
    self._force_outplace=_force_outplace
    self._module_class=_module_class
    
    for k,v in kwargs.items():
      setattr(self, k, v)

class TorchscriptToTorch:
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def __init__(
    self,
    map_location=None,
    _extra_files=None,
    **kwargs
  ):
    '''AUTOGENERATED by ``nbox``, some arguments might be missing, USE ``kwargs`` CAREFULLY!

    Load a :class:`ScriptModule` or :class:`ScriptFunction` previously
    saved with :func:`torch.jit.save <torch.jit.save>`

    All previously saved modules, no matter their device, are first loaded onto CPU,
    and then are moved to the devices they were saved from. If this fails (e.g.
    because the run time system doesn't have certain devices), an exception is
    raised.

    Args:
        f: a file-like object (has to implement read, readline, tell, and seek),
            or a string containing a file name
        map_location (string or torch.device): A simplified version of
            ``map_location`` in `torch.jit.save` used to dynamically remap
            storages to an alternative set of devices.
        _extra_files (dictionary of filename to content): The extra
            filenames given in the map would be loaded and their content
            would be stored in the provided map.

    Returns:
        A :class:`ScriptModule` object.

    Example:

    .. testcode::

        import torch
        import io

        torch.jit.load('scriptmodule.pt')

        # Load ScriptModule from io.BytesIO object
        with open('scriptmodule.pt', 'rb') as f:
            buffer = io.BytesIO(f.read())

        # Load all tensors to the original device
        torch.jit.load(buffer)

        # Load all tensors onto CPU, using a device
        buffer.seek(0)
        torch.jit.load(buffer, map_location=torch.device('cpu'))

        # Load all tensors onto CPU, using a string
        buffer.seek(0)
        torch.jit.load(buffer, map_location='cpu')

        # Load with extra files.
        extra_files = {'foo.txt': ''}  # values will be replaced with data
        torch.jit.load('scriptmodule.pt', _extra_files=extra_files)
        print(extra_files['foo.txt'])

    .. testoutput::
        :hide:

        ...

    .. testcleanup::

        import os
        os.remove("scriptmodule.pt")
    '''
    self.map_location=map_location
    self._extra_files=_extra_files
    
    for k,v in kwargs.items():
      setattr(self, k, v)

class TorchToOnnx:
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def __init__(
    self,
    model=None,
    args=None,
    export_params=True,
    verbose=False,
    input_names=None,
    output_names=None,
    operator_export_type=None,
    opset_version=None,
    do_constant_folding=True,
    dynamic_axes=None,
    keep_initializers_as_inputs=None,
    custom_opsets=None,
    export_modules_as_functions=False,
    **kwargs
  ):
    '''AUTOGENERATED by ``nbox``, some arguments might be missing, USE ``kwargs`` CAREFULLY!

    Exports a model into ONNX format. If ``model`` is not a
    :class:`torch.jit.ScriptModule` nor a :class:`torch.jit.ScriptFunction`, this runs
    ``model`` once in order to convert it to a TorchScript graph to be exported
    (the equivalent of :func:`torch.jit.trace`). Thus this has the same limited support
    for dynamic control flow as :func:`torch.jit.trace`.

    Args:
        model (torch.nn.Module, torch.jit.ScriptModule or torch.jit.ScriptFunction):
            the model to be exported.
        args (tuple or torch.Tensor):

            args can be structured either as:

            1. ONLY A TUPLE OF ARGUMENTS::

                args = (x, y, z)

            The tuple should contain model inputs such that ``model(*args)`` is a valid
            invocation of the model. Any non-Tensor arguments will be hard-coded into the
            exported model; any Tensor arguments will become inputs of the exported model,
            in the order they occur in the tuple.

            2. A TENSOR::

                args = torch.Tensor([1])

            This is equivalent to a 1-ary tuple of that Tensor.

            3. A TUPLE OF ARGUMENTS ENDING WITH A DICTIONARY OF NAMED ARGUMENTS::

                args = (x,
                        {'y': input_y,
                         'z': input_z})

            All but the last element of the tuple will be passed as non-keyword arguments,
            and named arguments will be set from the last element. If a named argument is
            not present in the dictionary, it is assigned the default value, or None if a
            default value is not provided.

            .. note::
                If a dictionary is the last element of the args tuple, it will be
                interpreted as containing named arguments. In order to pass a dict as the
                last non-keyword arg, provide an empty dict as the last element of the args
                tuple. For example, instead of::

                    torch.onnx.export(
                        model,
                        (x,
                         # WRONG: will be interpreted as named arguments
                         {y: z}),
                        "test.onnx.pb")

                Write::

                    torch.onnx.export(
                        model,
                        (x,
                         {y: z},
                         {}),
                        "test.onnx.pb")

        f: a file-like object (such that ``f.fileno()`` returns a file descriptor)
            or a string containing a file name.  A binary protocol buffer will be written
            to this file.
        export_params (bool, default True): if True, all parameters will
            be exported. Set this to False if you want to export an untrained model.
            In this case, the exported model will first take all of its parameters
            as arguments, with the ordering as specified by ``model.state_dict().values()``
        verbose (bool, default False): if True, prints a description of the
            model being exported to stdout. In addition, the final ONNX graph will include the
            field ``doc_string``` from the exported model which mentions the source code locations
            for ``model``.
        training (enum, default TrainingMode.EVAL):
            * ``TrainingMode.EVAL``: export the model in inference mode.
            * ``TrainingMode.PRESERVE``: export the model in inference mode if model.training is
              False and in training mode if model.training is True.
            * ``TrainingMode.TRAINING``: export the model in training mode. Disables optimizations
              which might interfere with training.
        input_names (list of str, default empty list): names to assign to the
            input nodes of the graph, in order.
        output_names (list of str, default empty list): names to assign to the
            output nodes of the graph, in order.
        operator_export_type (enum, default None):

            None usually means ``OperatorExportTypes.ONNX``.
            However if PyTorch was built with ``-DPYTORCH_ONNX_CAFFE2_BUNDLE``, None means
            ``OperatorExportTypes.ONNX_ATEN_FALLBACK``.

            * ``OperatorExportTypes.ONNX``: Export all ops as regular ONNX ops
              (in the default opset domain).
            * ``OperatorExportTypes.ONNX_FALLTHROUGH``: Try to convert all ops
              to standard ONNX ops in the default opset domain. If unable to do so
              (e.g. because support has not been added to convert a particular torch op to ONNX),
              fall back to exporting the op into a custom opset domain without conversion. Applies
              to `custom ops <https://pytorch.org/tutorials/advanced/torch_script_custom_ops.html>`_
              as well as ATen ops. For the exported model to be usable, the runtime must support
              these non-standard ops.
            * ``OperatorExportTypes.ONNX_ATEN``: All ATen ops (in the TorchScript namespace "aten")
              are exported as ATen ops (in opset domain "org.pytorch.aten").
              `ATen <https://pytorch.org/cppdocs/#aten>`_ is PyTorch's built-in tensor library, so
              this instructs the runtime to use PyTorch's implementation of these ops.

              .. warning::

                Models exported this way are probably runnable only by Caffe2.

              This may be useful if the numeric differences in implementations of operators are
              causing large differences in behavior between PyTorch and Caffe2 (which is more
              common on untrained models).

            * ``OperatorExportTypes.ONNX_ATEN_FALLBACK``: Try to export each ATen op
              (in the TorchScript namespace "aten") as a regular ONNX op. If we are unable to do so
              (e.g. because support has not been added to convert a particular torch op to ONNX),
              fall back to exporting an ATen op. See documentation on OperatorExportTypes.ONNX_ATEN for
              context.
              For example::

                graph(%0 : Float):
                  %3 : int = prim::Constant[value=0]()
                  # conversion unsupported
                  %4 : Float = aten::triu(%0, %3)
                  # conversion supported
                  %5 : Float = aten::mul(%4, %0)
                  return (%5)

              Assuming ``aten::triu`` is not supported in ONNX, this will be exported as::

                graph(%0 : Float):
                  %1 : Long() = onnx::Constant[value={0}]()
                  # not converted
                  %2 : Float = aten::ATen[operator="triu"](%0, %1)
                  # converted
                  %3 : Float = onnx::Mul(%2, %0)
                  return (%3)

              If PyTorch was built with Caffe2 (i.e. with ``BUILD_CAFFE2=1``), then
              Caffe2-specific behavior will be enabled, including special support
              for ops are produced by the modules described in
              `Quantization <https://pytorch.org/docs/stable/quantization.html>`_.

              .. warning::

                Models exported this way are probably runnable only by Caffe2.

        opset_version (int, default 9): The version of the
            `default (ai.onnx) opset <https://github.com/onnx/onnx/blob/master/docs/Operators.md>`_
            to target. Must be >= 7 and <= 15.
        do_constant_folding (bool, default True): Apply the constant-folding optimization.
            Constant-folding will replace some of the ops that have all constant inputs
            with pre-computed constant nodes.
        dynamic_axes (dict<string, dict<int, string>> or dict<string, list(int)>, default empty dict):

            By default the exported model will have the shapes of all input and output tensors
            set to exactly match those given in ``args``. To specify axes of tensors as
            dynamic (i.e. known only at run-time), set ``dynamic_axes`` to a dict with schema:

            * KEY (str): an input or output name. Each name must also be provided in ``input_names`` or
              ``output_names``.
            * VALUE (dict or list): If a dict, keys are axis indices and values are axis names. If a
              list, each element is an axis index.

            For example::

                class SumModule(torch.nn.Module):
                    def forward(self, x):
                        return torch.sum(x, dim=1)

                torch.onnx.export(SumModule(), (torch.ones(2, 2),), "onnx.pb",
                                  input_names=["x"], output_names=["sum"])

            Produces::

                input {
                  name: "x"
                  ...
                      shape {
                        dim {
                          dim_value: 2  # axis 0
                        }
                        dim {
                          dim_value: 2  # axis 1
                ...
                output {
                  name: "sum"
                  ...
                      shape {
                        dim {
                          dim_value: 2  # axis 0
                ...

            While::

                torch.onnx.export(SumModule(), (torch.ones(2, 2),), "onnx.pb",
                                  input_names=["x"], output_names=["sum"],
                                  dynamic_axes={
                                      # dict value: manually named axes
                                      "x": {0: "my_custom_axis_name"},
                                      # list value: automatic names
                                      "sum": [0],
                                  })

            Produces::

                input {
                  name: "x"
                  ...
                      shape {
                        dim {
                          dim_param: "my_custom_axis_name"  # axis 0
                        }
                        dim {
                          dim_value: 2  # axis 1
                ...
                output {
                  name: "sum"
                  ...
                      shape {
                        dim {
                          dim_param: "sum_dynamic_axes_1"  # axis 0
                ...

        keep_initializers_as_inputs (bool, default None): If True, all the
            initializers (typically corresponding to parameters) in the
            exported graph will also be added as inputs to the graph. If False,
            then initializers are not added as inputs to the graph, and only
            the non-parameter inputs are added as inputs.
            This may allow for better optimizations (e.g. constant folding) by
            backends/runtimes.

            If ``opset_version < 9``, initializers MUST be part of graph
            inputs and this argument will be ignored and the behavior will be
            equivalent to setting this argument to True.

            If None, then the behavior is chosen automatically as follows:

            * If ``operator_export_type=OperatorExportTypes.ONNX``, the behavior is equivalent
              to setting this argument to False.
            * Else, the behavior is equivalent to setting this argument to True.

        custom_opsets (dict<str, int>, default empty dict): A dict with schema:

            * KEY (str): opset domain name
            * VALUE (int): opset version

            If a custom opset is referenced by ``model`` but not mentioned in this dictionary,
            the opset version is set to 1. Only custom opset domain name and version should be
            indicated through this argument.

        export_modules_as_functions (bool or set of type of nn.Module, default False): Flag to enable
            exporting all ``nn.Module`` forward calls as local functions in ONNX. Or a set to indicate the
            particular types of modules to export as local functions in ONNX.
            This feature requires ``opset_version`` >= 15, otherwise the export will fail. This is because
            ``opset_version`` < 15 implies IR version < 8, which means no local function support.

            * ``False``(default): export ``nn.Module`` forward calls as fine grained nodes.
            * ``True``: export all ``nn.Module`` forward calls as local function nodes.
            * Set of type of nn.Module: export ``nn.Module`` forward calls as local function nodes,
              only if the type of the ``nn.Module`` is found in the set.

    Raises:
      CheckerError: If the ONNX checker detects an invalid ONNX graph. Will still export the
        model to the file ``f`` even if this is raised.
    '''
    self.model=model
    self.args=args
    self.export_params=export_params
    self.verbose=verbose
    self.input_names=input_names
    self.output_names=output_names
    self.operator_export_type=operator_export_type
    self.opset_version=opset_version
    self.do_constant_folding=do_constant_folding
    self.dynamic_axes=dynamic_axes
    self.keep_initializers_as_inputs=keep_initializers_as_inputs
    self.custom_opsets=custom_opsets
    self.export_modules_as_functions=export_modules_as_functions
    
    for k,v in kwargs.items():
      setattr(self, k, v)


class Framework_torch:
  _load_framework="torch"
  def _conditional(py_model):
    from .on_ml import torch_condition
    return torch_condition(py_model)
  
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def torch_to_torchscript(user_options: TorchToTorchscript, nbox_options: NboxOptions) -> Union[ModelSpec, Any]:
    '''
    Args:
      user_options (nbox.framework.ml.TorchToTorchscript): Variables to be filled by the user
      _nbox_options (nbox.framework.NboxOptions): Options to be used by NBX-Services
    Returns:
      when serialising the model returns ``ModelSpec`` else returns the ``model`` in ``nbox.Model``
    '''
    from .on_ml import torch_export_torchscript
    exported_time = Timestamp(); exported_time.GetCurrentTime()
    spec=ModelSpec(
      source=Framework(package="torch", method="Framework_torch.torch_to_torchscript"),
      target=Framework(package="torch", method="Framework_torch.torchscript_to_torch"),
      folder=nbox_options.folder,
      name=nbox_options.model_name,
      requirements=["-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch"], # contents inside requirements.txt file
      exported_time=exported_time,
      inputs=[],
      outputs=[],
    )
    spec=torch_export_torchscript(user_options, nbox_options, spec) # spec can be updated by the function
    return spec
  
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def torchscript_to_torch(user_options: TorchscriptToTorch, nbox_options: NboxOptions) -> Union[ModelSpec, Any]:
    '''
    Args:
      user_options (nbox.framework.ml.TorchscriptToTorch): Variables to be filled by the user
      _nbox_options (nbox.framework.NboxOptions): Options to be used by NBX-Services
    Returns:
      when serialising the model returns ``ModelSpec`` else returns the ``model`` in ``nbox.Model``
    '''
    from .on_ml import torch_import_torchscript
    exported_time = Timestamp(); exported_time.GetCurrentTime()
    spec=ModelSpec(
      source=Framework(package="torch", method="Framework_torch.torchscript_to_torch"),
      target=None, # this function has not been defined to throw any output
      folder=nbox_options.folder,
      name=nbox_options.model_name,
      requirements=["-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch"], # contents inside requirements.txt file
      exported_time=exported_time,
      inputs=[],
      outputs=[],
    )
    spec=torch_import_torchscript(user_options, nbox_options, spec) # spec can be updated by the function
    return spec
  
  # @isthere("-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch", soft=False)
  def torch_to_onnx(user_options: TorchToOnnx, nbox_options: NboxOptions) -> Union[ModelSpec, Any]:
    '''
    Args:
      user_options (nbox.framework.ml.TorchToOnnx): Variables to be filled by the user
      _nbox_options (nbox.framework.NboxOptions): Options to be used by NBX-Services
    Returns:
      when serialising the model returns ``ModelSpec`` else returns the ``model`` in ``nbox.Model``
    '''
    from .on_ml import torch_export_onnx
    exported_time = Timestamp(); exported_time.GetCurrentTime()
    spec=ModelSpec(
      source=Framework(package="torch", method="Framework_torch.torch_to_onnx"),
      target=Framework(package="onnx", method="Framework_onnx.onnx_runtime"),
      folder=nbox_options.folder,
      name=nbox_options.model_name,
      requirements=["-f https://download.pytorch.org/whl/cpu/torch_stable.html","torch"], # contents inside requirements.txt file
      exported_time=exported_time,
      inputs=[],
      outputs=[],
    )
    spec=torch_export_onnx(user_options, nbox_options, spec) # spec can be updated by the function
    return spec
  
  _METHODS={
    'torch_to_torchscript': (torch_to_torchscript, TorchToTorchscript),
    'torchscript_to_torch': (torchscript_to_torch, TorchscriptToTorch),
    'torch_to_onnx': (torch_to_onnx, TorchToOnnx),
    }


# End of file


"""
gRPC + Protobuf = ♥️
"""

from grpc import RpcError
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.json_format import MessageToDict, MessageToJson, ParseDict

from .utils import logger

def rpc(stub: callable, message, err_msg: str, raise_on_error: bool = True):
  """convienience function for calling a gRPC stub"""
  try:
    resp = stub(message)
    raise_on_error = False
  except RpcError as e:
    logger.error(err_msg)
    # err_ = loads(e.debug_error_string())
    # if "value" in err_:
    #   if int(err_["value"]) > 500:
    #     logger.error("There is something wrong from our side. Your files are safe on your local machine.")
    #   elif int(err_["value"]) > 400:
    #     logger.error("There is something wrong in nbox. Raise an issue on github: https://github.com/NimbleBoxAI/nbox/issues")
    # logger.error(err_["grpc_message"])
    logger.error(e)
  else:
    return resp
  if raise_on_error:
    raise RpcError("NBX-RPC error, see above for details")

def streaming_rpc(stub: callable, message, err_msg: str, raise_on_error: bool = True):
  """convienience function for streaming from a gRPC stub"""
  try:
    data_iter = stub(message)
    for data in data_iter:
      yield data
  except RpcError as e:
    logger.error(err_msg)
    logger.error(e.details())
    if raise_on_error:
      raise e

def message_to_json(message):
  """convert message to json"""
  return MessageToJson(
    message=message,
    including_default_value_fields=True,
    preserving_proto_field_name=True,
    indent=2,
    sort_keys=False,
    use_integers_for_enums=True,
    float_precision=4
  )

def message_to_dict(message):
  """convert message to dict"""
  return MessageToDict(
    message = message,
    including_default_value_fields=True,
    preserving_proto_field_name=True,
    use_integers_for_enums=True,
    float_precision=4
  )

def dict_to_message(dict, message):
  """load dict into message"""
  ParseDict(dict, message)
  return message


def get_current_timestamp():
  ts = Timestamp()
  ts.GetCurrentTime()
  return ts



"""
Deprecated: `nbox.nbxlib.ops` is deprecated and will be removed in a future release, use `nbox.lib` instead.

MIT-License code for all operators that are open sourced."""

from nbox.lib.arch import *
from nbox.lib.aws import *
from nbox.lib.comms import *
from nbox.lib.demo import *
from nbox.lib.nbx_instances import *
from nbox.lib.shell import *



"""
``Operators`` is how you write NBX-Jobs. If you are familiar with pytorch, then usage is
exactly same, for others here's a quick recap:

.. code-block:: python

  class MyOperator(Operator):
    def __init__(self, a: int, b: str):
      super().__init__()
      self.a: int = a
      self.b: Operator = MyOtherOperator(b) # nested calling
    
    def forward(self, x: int) -> int:
      y = self.a + x
      y = self.b(y) + x # nested calling
      return y

  job = MyOperator(1, "hello") # define once
  res = job(2)                 # use like python, screw DAGs

If you want to use deploy `nbox.Jobs <nbox.jobs.html>`_ is a better documentation.


Engineering
-----------

Fundamentally operators act as a wrapper on user code, sometime abstracting away functions
by breaking them into ``__init__``s and ``forward``s. But this is a simpler way to wrap
user function than letting users wrap their own function. It is easy to get false positives,
and so we explicitly expand things in two. These operators are like ``torch.nn.Modules``
spiritually as well because modules manage the underlying weights and operators manage the
underlying user logic.

Operators are combination of several subsystems that are all added in the same class, though
certainly if we come up with that high abstraction we will refactor this:

#. tree: All operators are really treated like a tree meaning that the execution is nested\
    and the order of execution is determined by the order of the operators in the tree. DAGs\
    are fundamentally just trees with some nodes spun togeather, to execute only once.
#. deploy, ...: All the services in NBX-Jobs.
#. get_nbx_flow: which is the static code analysis system to understand true user intent and\
    if possible (and permission of the user) optimise the logic.
"""
# Some parts of the code are based on the pytorch nn.Module class
# pytorch license: https://github.com/pytorch/pytorch/blob/master/LICENSE
# modifications: research@nimblebox.ai 2022

import re
import os
import jinja2
import zipfile
import fnmatch
from hashlib import sha256
from tempfile import gettempdir
from collections import OrderedDict
from typing import Callable, Iterable, List

from . import utils as U
from .utils import logger
from .init import nbox_ws_v1
from .network import deploy_job, Schedule, deploy_serving
from .framework.on_functions import get_nbx_flow
from .framework import AirflowMixin, PrefectMixin, LuigiMixin
from .hyperloop.job_pb2 import NBXAuthInfo, Job as JobProto, Resource
from .hyperloop.dag_pb2 import DAG, Node, RunStatus
from .hyperloop.serve_pb2 import Serving
from .nbxlib.tracer import Tracer
from .messages import get_current_timestamp
from .sub_utils.latency import log_latency
from .version import __version__
from nbox.auth import secret

def get_fn_base_models(fn) -> List[str]:
  import ast
  if type(fn) == str:
    fn = ast.parse(fn)
  
  for node in fn.body:
    if type(node) == ast.FunctionDef:
      args: ast.arguments = node.args
      all_args = [] # tuple of name, type (if any) and default value
      for arg in args.args[1:]: # first one is always self
        all_args.append(arg.arg)
      if args.kwonlyargs:
        for arg in args.kwonlyargs:
          all_args.append(arg.arg)
      if args.vararg:
        all_args.append(args.vararg.arg)
      if args.kwarg:
        all_args.append(args.kwarg.arg)
  
  strings = [f"{x}: Any = None" for x in all_args]
  return strings

class Operator():
  _version: int = 1 # always try to keep this an i32

  def __init__(self) -> None:
    """Create an operator, which abstracts your code into sharable, bulding blocks which
    can then deployed on either NBX-Jobs or NBX-Deploy.
    
    Usage:

    .. code-block:: python
      
      class MyOperator(Operator):
        def __init__(self, ...):
          ... # initialisation of job happens here

          # use prebuilt operators to define the entire process
          from .nbxlib.ops import Shell
          self.download_binary = Shell("wget https://nbox.ai/{hash}")

          # keep a library of organisation wide operators
          from .nbx_internal.operators import TrainGPT
          self.train_model: Operator = TrainGPT(...)

        def forward(self, ...):
          # pass any inputs you want at runtime
          ... # execute code in any arbitrary order, free from DAGs

          self.download_binary(hash="my_binary") # pass relevant data at runtime
          self.train_model() # run any operation and get full visibility on platform

      # to convert operator is 
      job: Operator = MyOperator(...)
    """
    self._operators = OrderedDict() # {name: operator}
    self._op_trace = []
    self._tracer: Tracer = None

  def __remote_init__(self):
    """User can overwrite this function, this will be called only when running on remote.
    This helps in with things like creating the models can caching them in self, instead
    of ``lru_cache`` in forward."""
    pass

  # mixin/

  to_airflow_operator = AirflowMixin.to_airflow_operator
  to_airflow_dag = AirflowMixin.to_airflow_dag
  from_airflow_operator = classmethod(AirflowMixin.from_airflow_operator)
  from_airflow_dag = classmethod(AirflowMixin.from_airflow_dag)
  to_prefect_task = PrefectMixin.to_prefect_task
  to_prefect_flow = PrefectMixin.to_prefect_flow
  from_prefect_task = classmethod(PrefectMixin.from_prefect_task)
  from_prefect_flow = classmethod(PrefectMixin.from_prefect_flow)
  to_luigi_task = LuigiMixin.to_luigi_task
  to_luigi_flow = LuigiMixin.to_luigi_flow
  from_luigi_task = classmethod(LuigiMixin.from_luigi_task)
  from_luigi_flow = classmethod(LuigiMixin.from_luigi_flow)

  # /mixin

  # information passing/

  def __repr__(self):
    # from torch.nn.Module
    def _addindent(s_, numSpaces):
      s = s_.split('\n')
      # don't do anything for single-line stuff
      if len(s) == 1:
        return s_
      first = s.pop(0)
      s = [(numSpaces * ' ') + line for line in s]
      s = '\n'.join(s)
      s = first + '\n' + s
      return s

    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    extra_repr = ""
    # empty string will be split into list ['']
    if extra_repr:
      extra_lines = extra_repr.split('\n')
    child_lines = []
    for key, module in self._operators.items():
      mod_str = repr(module)
      mod_str = _addindent(mod_str, 2)
      child_lines.append('(' + key + '): ' + mod_str)
    lines = extra_lines + child_lines

    main_str = self.__class__.__name__ + '('
    if lines:
      # simple one-liner info, which most builtin Modules will use
      if len(extra_lines) == 1 and not child_lines:
        main_str += extra_lines[0]
      else:
        main_str += '\n  ' + '\n  '.join(lines) + '\n'

    main_str += ')'
    return main_str

  def __setattr__(self, key, value: 'Operator'):
    obj = getattr(self, key, None)
    if key != "forward" and obj is not None and callable(obj):
      raise AttributeError(f"cannot assign {key} as it is already a method")
    if isinstance(value, Operator):
      if not "_operators" in self.__dict__:
        raise AttributeError("cannot assign operator before Operator.__init__() call")
      if key in self.__dict__ and key not in self._operators:
        raise KeyError(f"attribute '{key}' already exists")
      self._operators[key] = value
    self.__dict__[key] = value

  def propagate(self, **kwargs):
    """Set kwargs for each child in the Operator"""
    for c in self.children:
      c.propagate(**kwargs)
    for k, v in kwargs.items():
      setattr(self, k, v)

  def thaw(self, job: JobProto):
    """Load JobProto into this Operator"""
    nodes = job.dag.flowchart.nodes
    edges = job.dag.flowchart.edges
    for _id, node in nodes.items():
      name = node.name
      if name.startswith("self."):
        name = name[5:]
      if hasattr(self, name):
        op: 'Operator' = getattr(self, name)
        op.propagate(
          node = node,
          source_edges = list(filter(
            lambda x: edges[x].target == node.id, edges.keys()
          ))
        )
  
  # /information passing

  # properties/

  def operators(self):
    r"""Returns an iterator over all operators in the job."""
    for _, module in self.named_operators():
      yield module

  def named_operators(self, memo = None, prefix: str = '', remove_duplicate: bool = True):
    r"""Returns an iterator over all modules in the network, yielding both the name of the module
    as well as the module itself."""
    if memo is None:
      memo = set()
    if self not in memo:
      if remove_duplicate:
        memo.add(self)
      yield prefix, self
      for name, module in self._operators.items():
        if module is None:
          continue
        submodule_prefix = prefix + ('.' if prefix else '') + name
        for m in module.named_operators(memo, submodule_prefix, remove_duplicate):
          yield m

  @property
  def children(self) -> Iterable['Operator']:
    """Get children of the operator."""
    return self._operators.values()

  # user can manually define inputs if they want, avoid this user if you don't know how this system works

  _inputs = []
  @property
  def inputs(self):
    import inspect
    args = inspect.getfullargspec(self.forward).args
    try:
      args.remove('self')
    except:
      pass
    if self._inputs:
      args += self._inputs
    return args

  # /properties

  def forward(self):
    raise NotImplementedError("User must implement forward()")

  def _register_forward(self, python_callable: Callable):
    """convienience method to register a forward method"""
    self.forward = python_callable

  # define these here, it's okay to waste some memory
  node = Node()
  source_edges: List[Node] = None

  def __call__(self, *args, **kwargs):
    # Type Checking and create input dicts
    inputs = self.inputs
    len_inputs = len(args) + len(kwargs)
    if len_inputs > len(inputs):
      raise ValueError(f"Number of arguments ({len(inputs)}) does not match number of inputs ({len_inputs})")
    elif len_inputs < len(args):
      raise ValueError(f"Need at least arguments ({len(args)}) but got ({len_inputs})")

    with log_latency(f"{self.__class__.__name__}-prestep"):
      input_dict = {}
      for i, arg in enumerate(args):
        input_dict[self.inputs[i]] = arg
      for key, value in kwargs.items():
        if key in inputs:
          input_dict[key] = value

      logger.debug(f"Calling operator '{self.__class__.__name__}': {self.node.id}")
      _ts = get_current_timestamp()
      self.node.run_status.CopyFrom(RunStatus(start = _ts, inputs = {k: str(type(v)) for k, v in input_dict.items()}))
      if self._tracer != None:
        self._tracer(self.node)

    # ---- USER SEPERATION BOUNDARY ---- #

    with log_latency(f"{self.__class__.__name__}-forward"):
      out = self.forward(*args, **kwargs)

    # ---- USER SEPERATION BOUNDARY ---- #
    with log_latency(f"{self.__class__.__name__}-poststep"):
      outputs = {}
      if out is None:
        outputs = {"out_0": str(type(None))}
      elif isinstance(out, dict):
        outputs = {k: str(type(v)) for k, v in out.items()}
      elif isinstance(out, (list, tuple)):
        outputs = {f"out_{i}": str(type(v)) for i, v in enumerate(out)}
      else:
        outputs = {"out_0": str(type(out))}

      logger.debug(f"Ending operator '{self.__class__.__name__}': {self.node.id}")
      _ts = get_current_timestamp()
      self.node.run_status.MergeFrom(RunStatus(end = _ts, outputs = outputs,))
      if self._tracer != None:
        self._tracer(self.node)

    return out

  # nbx/

  def deploy(
    self,
    init_folder: str,
    job_id_or_name: str,
    workspace_id: str = None,
    schedule: Schedule = None,
    resource: Resource = None,
    *,
    _unittest = False
  ):
    """Deploy this job on NBX-Jobs.

    DO NOT CALL THIS DIRECTLY, use `nbx jobs new` CLI command instead.

    Args:
        init_folder (str, optional): Name the folder to zip
        job_id_or_name (Union[str, int], optional): Name or ID of the job
        workspace_id (str): Workspace ID to deploy to, if not specified, will use the personal workspace
        schedule (Schedule, optional): If ``None`` will run only once, else will schedule the job
        cache_dir (str, optional): Folder where to put the zipped file, if ``None`` will be ``tempdir``
    Returns:
        Job: Job object
    """
    # get stub
    if workspace_id == None:
      stub_all_jobs = nbox_ws_v1.user.jobs
    else:
      stub_all_jobs = nbox_ws_v1.workspace.u(workspace_id).jobs

    all_jobs = stub_all_jobs()["data"]

    jobs = list(filter(lambda x: x["job_id"] == job_id_or_name or x["name"] == job_id_or_name, all_jobs))
    if len(jobs) == 0:
      logger.info(f"No Job found with ID or name: {job_id_or_name}, will create a new one")
      job_name =  job_id_or_name
      job_id = None
    elif len(jobs) > 1:
      raise ValueError(f"Multiple jobs found for '{job_id_or_name}', try passing ID")
    else:
      logger.info(f"Found job with ID or name: {job_id_or_name}, will update it")
      data = jobs[0]
      job_name = data["name"]
      job_id = data["job_id"]
    
    logger.info(f"Job name: {job_name}")
    logger.info(f"Job ID: {job_id}")

    # check if this is a valid folder or not
    if not os.path.exists(init_folder) or not os.path.isdir(init_folder):
      raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")
    if os.path.isdir(init_folder):
      os.chdir(init_folder)
      if not os.path.exists("./exe.py"):
        raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")
    else:
      raise ValueError(f"Incorrect project at path: '{init_folder}'! nbox jobs new <name>")

    # flowchart-alpha
    dag: DAG = get_nbx_flow(self.forward)
    for _id, node in dag.flowchart.nodes.items():
      name = node.name
      if name.startswith("self."):
        name = name[5:]
      operator_name = "CodeBlock" # default
      cls_item = getattr(self, name, None)
      if cls_item and cls_item.__class__.__base__ == Operator:
        operator_name = cls_item.__class__.__name__
      node.operator = operator_name

    if schedule != None:
      logger.debug(f"Schedule: {schedule}")

    _starts = get_current_timestamp()
    job_proto = JobProto(
      id = job_id,
      name = job_name if job_name else U.get_random_name(True).split("-")[0],
      created_at = _starts,
      auth_info = NBXAuthInfo(
        username = secret.get("username"),
        workspace_id = workspace_id,
      ),
      schedule = schedule.get_message() if schedule != None else None,
      dag = dag,
      resource = Resource(
        cpu = "100m",         # 100mCPU
        memory = "200Mi",     # MiB
        disk_size = "1Gi",    # GiB
      ) if resource == None else resource,
    )

    with open(U.join(init_folder, "job_proto.msg"), "wb") as f:
      f.write(job_proto.SerializeToString())

    # # create the runner stub: in case of jobs it will be python3 exe.py run
    # py_data = dict(
    #   import_string_nbox = "from nbox.network import Schedule" if scheduled else None,
    #   job_id_or_name = job_id_or_name,
    #   workspace_id = workspace_id,
    #   scheduled = schedule,
    #   project_name = project_name,
    #   created_time = created_time,
    # )
    # py_f_data = {k:v for k,v in py_data.items() if v is not None}

    # path = U.join(assets, "job_nbx.jinja")
    # with open(path, "r") as f, open("exe.py", "w") as f2:
    #   f2.write(jinja2.Template(f.read()).render(**py_f_data))

    # zip all the files folder
    all_f = U.get_files_in_folder(init_folder)

    # find a .nboxignore file and ignore items in it
    to_ignore_pat = []
    to_ignore_folder = []
    for f in all_f:
      if f.split("/")[-1] == ".nboxignore":
        # we have this file
        with open(f, "r") as _f:
          for pat in _f:
            pat = pat.strip()
            if pat.endswith("/"):
              to_ignore_folder.append(pat)
            else:
              to_ignore_pat.append(pat)
        break

    # two different lists for convinience
    to_remove = []
    for ignore in to_ignore_pat:
      x = fnmatch.filter(all_f, ignore)
      to_remove.extend(x)
    to_remove_folder = []
    for ignore in to_ignore_folder:
      for f in all_f:
        if re.search(ignore, f):
          to_remove_folder.append(f)
    to_remove += to_remove_folder

    all_f = [x for x in all_f if x not in to_remove]
    logger.info(f"Will zip {len(all_f)} files")

    for f in all_f:
      hash_ = sha256()
      with open(f, "rb") as f:
        for c in iter(lambda: f.read(2 ** 20), b""):
          hash_.update(c)
    hash_ = hash_.hexdigest()
    logger.info(f"SHA256 ( {init_folder} ): {hash_}")

    zip_path = U.join(gettempdir(), f"project-{hash_}.nbox")
    logger.info(f"Packing project to '{zip_path}'")
    with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zip_file:
      abspath_init_folder = os.path.abspath(init_folder)
      for f in all_f:
        arcname = f[len(abspath_init_folder)+1:]
        logger.debug(f"Zipping {f} => {arcname}")
        zip_file.write(f, arcname = arcname)

    if _unittest:
      return job_proto
  
    return deploy_job(zip_path = zip_path, job_proto = job_proto)

  def serve(
    self,
    init_folder: str,
    deployment_id_or_name: str,
    workspace_id: str = None,
    resource: Resource = None,
    wait_for_deployment: bool = True,
    *,
    _unittest = False,
  ):
    """Serve your operator as an API endpoint.
    
    DO NOT CALL THIS DIRECTLY, use `nbx serve new` CLI command instead.

    EXPERIMENTAL: can break anytime
    """
    init_folder = os.path.abspath(init_folder)

    # raise NotImplementedError(f"In-progress will be released in v1.0.0 (currently {__version__})")
    if workspace_id == None:
      stub_all_depl = nbox_ws_v1.user.deployments
    else:
      stub_all_depl = nbox_ws_v1.workspace.u(workspace_id).deployments
    logger.debug(f"deployments stub: {stub_all_depl}")

    # filter and get "id" and "name"
    deployments = list(filter(
      lambda x: x["deployment_id"] == deployment_id_or_name or x["deployment_name"] == deployment_id_or_name,
      stub_all_depl()["data"]
    ))
    if len(deployments) == 0:
      logger.warning(f"No deployment found with id '{deployment_id_or_name}', creating new with name!")
      deployment_id = None
      deployment_name = deployment_id_or_name
    elif len(deployments) > 1:
      raise ValueError(f"Multiple deployments found for '{deployment_id_or_name}', try passing ID")
    else:
      data = deployments[0]
      deployment_id = data["deployment_id"]
      deployment_name = data["deployment_name"]

    logger.info(f"Deployment name: {deployment_name}")
    logger.info(f"Deployment ID: {deployment_id}")

    # create a serve proto and use that to serve the model using NBX-Infra
    _starts = get_current_timestamp()
    serving_proto = Serving(
      name = deployment_name,
      created_at = _starts,
      resource = Resource(
        cpu = "100m",         # 100mCPU
        memory = "200Mi",     # MiB
        disk_size = "1Gi",    # GiB
      ) if resource == None else resource
    )

    if workspace_id != None:
      serving_proto.workspace_id = workspace_id
    if deployment_id != None:
      serving_proto.id = deployment_id

    # create the runner stub: in case of jobs it will be python3 exe.py run
    import inspect
    from textwrap import dedent

    forward_code = inspect.getsource(self.forward)
    strings = get_fn_base_models(dedent(forward_code))

    py_data = dict(
      project_name = os.path.split(init_folder)[-1],
      created_time = _starts.ToDatetime().strftime("%Y-%m-%d %H:%M:%S"),
      email_id = None,
      operator_name = self.__class__.__name__,
      base_model_strings = strings
    )
    py_f_data = {k:v for k,v in py_data.items() if v is not None}
    assets = U.join(U.folder(__file__), "assets")
    path = U.join(assets, "job_serve.jinja")
    server_path = U.join(init_folder, f"server.py")
    with open(path, "r") as f, open(server_path, "w") as f2:
      f2.write(jinja2.Template(f.read()).render(**py_f_data))

    # zip tje folder
    zip_path = self.zip(init_folder)

    if _unittest:
      return serving_proto

    return deploy_serving(
      export_model_path = zip_path,
      stub_all_depl = stub_all_depl,
      wait_for_deployment = wait_for_deployment
    )

  @staticmethod
  def zip(init_folder: str):
    all_f = [os.path.join(init_folder, x) for x in U.get_files_in_folder(init_folder)]
    zip_arcnames = [x[len(init_folder)+1:] for x in all_f] # arcnames are the relative names inside the zip
    for f in all_f:
      hash_ = sha256()
      with open(f, "rb") as f:
        for c in iter(lambda: f.read(2 ** 20), b""):
          hash_.update(c)
    hash_ = hash_.hexdigest()
    logger.info(f"SHA256 ( {init_folder} ): {hash_}")

    zip_path = U.join(gettempdir(), f"project-{hash_}.nbox")
    logger.info(f"Packing to path: {zip_path}")
    with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zip_file:
      for f, arcname in zip(all_f, zip_arcnames):
        zip_file.write(filename = f, arcname = arcname)

    return zip_file

  @staticmethod
  def unzip(zip_path: str, dest_folder: str):
    """Unzip a zip file to a folder.
    """
    with zipfile.ZipFile(zip_path, 'r') as zip_file:
      zip_file.extractall(dest_folder)

  # /nbx



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class DAG(google.protobuf.message.Message):
    """DAG is the total set of operations to execute, though not technically
    DAGs since there can be recursion or anything that python can do. It
    is supposed to be our understanding of the code with our own internal
    IR.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class SymbolsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___Node: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___Node] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    SYMBOLS_FIELD_NUMBER: builtins.int
    FLOWCHART_FIELD_NUMBER: builtins.int
    @property
    def symbols(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___Node]:
        """symbols is a dict with name keys and nodes as values"""
        pass
    @property
    def flowchart(self) -> global___Flowchart: ...
    def __init__(self,
        *,
        symbols: typing.Optional[typing.Mapping[typing.Text, global___Node]] = ...,
        flowchart: typing.Optional[global___Flowchart] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["flowchart",b"flowchart"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["flowchart",b"flowchart","symbols",b"symbols"]) -> None: ...
global___DAG = DAG

class Flowchart(google.protobuf.message.Message):
    """Flowchart represents the data flow with nodes and edges, this can then
    be optimised for faster more performant execution. With nodes and edges
    we can reconstruct and play with any kind of arbitrary code.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class NodesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___Node: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___Node] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class EdgesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___Edge: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___Edge] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    NODES_FIELD_NUMBER: builtins.int
    EDGES_FIELD_NUMBER: builtins.int
    @property
    def nodes(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___Node]:
        """node.id: node dict; map because easier to fetch and update things"""
        pass
    @property
    def edges(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___Edge]:
        """edge.id: edge dict; map because easier to fetch and update things"""
        pass
    def __init__(self,
        *,
        nodes: typing.Optional[typing.Mapping[typing.Text, global___Node]] = ...,
        edges: typing.Optional[typing.Mapping[typing.Text, global___Edge]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["edges",b"edges","nodes",b"nodes"]) -> None: ...
global___Flowchart = Flowchart

class Edge(google.protobuf.message.Message):
    """The edges are just connectors and should not contain any data of its own
    nbox_string can be used to tell about the edge in our IR
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _EdgeType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _EdgeTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Edge._EdgeType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        EXECUTION_ORDER: Edge._EdgeType.ValueType  # 0
        BRANCHING: Edge._EdgeType.ValueType  # 1
        LOOP: Edge._EdgeType.ValueType  # 2
    class EdgeType(_EdgeType, metaclass=_EdgeTypeEnumTypeWrapper):
        """execution order: default
        branching: conditionals or parallels
        loop edge: for or while loop things
        """
        pass

    EXECUTION_ORDER: Edge.EdgeType.ValueType  # 0
    BRANCHING: Edge.EdgeType.ValueType  # 1
    LOOP: Edge.EdgeType.ValueType  # 2

    ID_FIELD_NUMBER: builtins.int
    SOURCE_FIELD_NUMBER: builtins.int
    TARGET_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    NBOX_STRING_FIELD_NUMBER: builtins.int
    id: typing.Text
    """edge id should be a combination of source and target"""

    source: typing.Text
    """name/id of the source"""

    target: typing.Text
    """name/id of the target"""

    type: global___Edge.EdgeType.ValueType
    """what type of edge is this, it can tell if it's a conditional
    edge or a loop edge or something
    """

    nbox_string: typing.Text
    def __init__(self,
        *,
        id: typing.Text = ...,
        source: typing.Text = ...,
        target: typing.Text = ...,
        type: global___Edge.EdgeType.ValueType = ...,
        nbox_string: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["id",b"id","nbox_string",b"nbox_string","source",b"source","target",b"target","type",b"type"]) -> None: ...
global___Edge = Edge

class RunStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class InputsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class OutputsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    START_FIELD_NUMBER: builtins.int
    END_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    @property
    def start(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """start datetime of this node"""
        pass
    @property
    def end(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """end datetime of this node"""
        pass
    @property
    def inputs(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """name of the inputs and optionally some value against them"""
        pass
    @property
    def outputs(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """name of the outputs with some string against it"""
        pass
    def __init__(self,
        *,
        start: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        end: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        inputs: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        outputs: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["end",b"end","start",b"start"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["end",b"end","inputs",b"inputs","outputs",b"outputs","start",b"start"]) -> None: ...
global___RunStatus = RunStatus

class Code(google.protobuf.message.Message):
    """Code mostly has code related information"""
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class InputsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class OutputsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        value: typing.Text
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    CODE_FIELD_NUMBER: builtins.int
    NBOX_STRING_FIELD_NUMBER: builtins.int
    LINENO_FIELD_NUMBER: builtins.int
    COL_OFFSET_FIELD_NUMBER: builtins.int
    END_LINENO_FIELD_NUMBER: builtins.int
    END_COL_OFFSET_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    CONDITIONS_FIELD_NUMBER: builtins.int
    name: typing.Text
    """name of the node"""

    code: typing.Text
    """code string for this node"""

    nbox_string: typing.Text
    lineno: builtins.int
    """these are code location elements"""

    col_offset: builtins.int
    end_lineno: builtins.int
    end_col_offset: builtins.int
    @property
    def inputs(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """name of the inputs and optionally some value against them"""
        pass
    @property
    def outputs(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """name of the outputs with some string against it"""
        pass
    @property
    def conditions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Code]:
        """when this is if-conditional it can have many different nodes"""
        pass
    def __init__(self,
        *,
        name: typing.Text = ...,
        code: typing.Text = ...,
        nbox_string: typing.Text = ...,
        lineno: builtins.int = ...,
        col_offset: builtins.int = ...,
        end_lineno: builtins.int = ...,
        end_col_offset: builtins.int = ...,
        inputs: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        outputs: typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        conditions: typing.Optional[typing.Iterable[global___Code]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["code",b"code","col_offset",b"col_offset","conditions",b"conditions","end_col_offset",b"end_col_offset","end_lineno",b"end_lineno","inputs",b"inputs","lineno",b"lineno","name",b"name","nbox_string",b"nbox_string","outputs",b"outputs"]) -> None: ...
global___Code = Code

class Node(google.protobuf.message.Message):
    """Nodes contain the real-deal, these beast of a data structure has
    logic for processing some inputs, they are the servers and edges
    are just conne
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class _NodeType:
        ValueType = typing.NewType('ValueType', builtins.int)
        V: typing_extensions.TypeAlias = ValueType
    class _NodeTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Node._NodeType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        CODE: Node._NodeType.ValueType  # 0
        """code-block, this default and selected when code cannot be understood"""

        OP: Node._NodeType.ValueType  # 1
        """this is when the code is an operation and to be considered"""

        LOOP: Node._NodeType.ValueType  # 2
        """loop node can be a while or for loop"""

        BRANCHING: Node._NodeType.ValueType  # 3
        """code for parallelisation or conditional"""

        RETURN: Node._NodeType.ValueType  # 4
        """this code is returning something, defined for convinience"""

    class NodeType(_NodeType, metaclass=_NodeTypeEnumTypeWrapper):
        """NodeTypes can be of a few types such as"""
        pass

    CODE: Node.NodeType.ValueType  # 0
    """code-block, this default and selected when code cannot be understood"""

    OP: Node.NodeType.ValueType  # 1
    """this is when the code is an operation and to be considered"""

    LOOP: Node.NodeType.ValueType  # 2
    """loop node can be a while or for loop"""

    BRANCHING: Node.NodeType.ValueType  # 3
    """code for parallelisation or conditional"""

    RETURN: Node.NodeType.ValueType  # 4
    """this code is returning something, defined for convinience"""


    ID_FIELD_NUMBER: builtins.int
    EXECUTION_INDEX_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    OPERATOR_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    INFO_FIELD_NUMBER: builtins.int
    NBOX_STRING_FIELD_NUMBER: builtins.int
    RUN_STATUS_FIELD_NUMBER: builtins.int
    id: typing.Text
    """uuid4 id of the Node object"""

    execution_index: builtins.int
    """since the code during execution can be actually run out of order
    we need to store the order in which it would have been executed
    on the users single threaded GIL python for best experience. Thus
    adding execution_index which tracks this.
    """

    name: typing.Text
    """name of the node"""

    operator: typing.Text
    """operator is the name of the Operator class that is this node, it is
    to understanding what are the things people want to do since these
    names are given by the user
    """

    type: global___Node.NodeType.ValueType
    """Type of nodes"""

    @property
    def info(self) -> global___Code:
        """Code contains the code related information"""
        pass
    nbox_string: typing.Text
    @property
    def run_status(self) -> global___RunStatus:
        """define the run status"""
        pass
    def __init__(self,
        *,
        id: typing.Text = ...,
        execution_index: builtins.int = ...,
        name: typing.Text = ...,
        operator: typing.Text = ...,
        type: global___Node.NodeType.ValueType = ...,
        info: typing.Optional[global___Code] = ...,
        nbox_string: typing.Text = ...,
        run_status: typing.Optional[global___RunStatus] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["info",b"info","run_status",b"run_status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["execution_index",b"execution_index","id",b"id","info",b"info","name",b"name","nbox_string",b"nbox_string","operator",b"operator","run_status",b"run_status","type",b"type"]) -> None: ...
global___Node = Node
